#!/usr/bin/env python3

import subprocess
import os
import argparse
import platform

package_directory = os.path.dirname(os.path.abspath(__file__))

from fretraj import metadata
from fretraj import _LabelLib_found
from fretraj import _nglview_found
from fretraj import _burst_module_available
from fretraj import _jupyter_module_available
from fretraj import cloud
import mdtraj as md


def parseCmd():
    """Parse the command line to get the input PDB file and the dye parameters

    Returns
    -------
    tuple of str
        tuple containing the filenames of the input PDB and the parameter file, the labeling site identifier
        and the filename of the output ACV, i.e. (pdb_input, param_file, site, pdb_output)
    """
    parser = argparse.ArgumentParser(
        description="compute accessible-contact clouds for an MD trajectory or a given PDB structure"
    )
    parser.add_argument("--version", action="version", version="%(prog)s " + str(metadata["Version"]))
    parser.add_argument(
        "--path", action="version", version=f"package directory: {package_directory}", help="Show package directory"
    )
    parser.add_argument("-i", "--input", help="Input PDB structure (.pdb)", required=True)
    parser.add_argument("-p", "--parameters", help="Parameter file for labels (.json)", required=True)
    parser.add_argument("-s", "--site", help="reference identifier for the labeling position (string)", required=True)
    parser.add_argument(
        "-o", "--output", help="Output file of accessible contact volume (.pdb, .xyz, .dx)", required=False
    )
    parser.add_argument(
        "--show-config",
        action="version",
        version=f"labellib: {_LabelLib_found} | nglview: {_nglview_found} | burst submodule: {_burst_module_available} | jupyter submodule: {_jupyter_module_available}",
        help="Show installed libraries and submodules",
    )
    args = parser.parse_args()
    in_filePDB = args.input
    param_fileJSON = args.parameters
    out_fileACV = args.output
    site = args.site
    return (in_filePDB, param_fileJSON, site, out_fileACV)


def main():
    """Calcualting an ACV from the command line

    Notes
    -----
    Run with:
    >>> fretraj -i <input pdb> -p <label parameter file (.json)> -s <label position identifier>
    """
    in_filePDB, param_fileJSON, site, out_fileACV = parseCmd()
    labels = cloud.labeling_params(param_fileJSON)
    struct = md.load_pdb(in_filePDB)
    print("Calculating accessible-contact volume...")
    av = cloud.Volume(struct, site, labels)
    if not out_fileACV:
        out_fileACV = "ACV.pdb"
    _, out_fileextACV = os.path.splitext(out_fileACV)
    out_path = os.path.dirname(os.path.abspath(out_fileACV))
    av.save_acv(out_fileACV, format=out_fileextACV[1:])
    print(f"{out_fileACV} saved to {out_path}")


def pymol_vis():
    """Visualize a PDB file or trajectory in PyMOL

    Notes
    -----
    Run with:
    >>> pymol_vis -c <structure file (.gro/.pdb)> -x <xtc trajectory> (optional) -v <pymol visualization file> (optional) -s <start,stop,stride (default 1,-1,1)>
    """
    if platform.system() != "Windows":
        subprocess.call(os.path.join(package_directory, "skripts", "pymol_vis.sh"))
    else:
        print("pymol_vis is only available for Unix operating systems")


def vmd_vis():
    """Visualize a PDB file or trajectory in VMD

    Notes
    -----
    Run with:
    >>> vmd_vis -c <structure file (.gro/.pdb)> -x <xtc trajectory> (optional) -v <vmd visualization file> (optional) -s <start,stop,stride (default 1,-1,1)>
    """
    if platform.system() != "Windows":
        subprocess.call(os.path.join(package_directory, "skripts", "vmd_vis.sh"))
    else:
        print("vmd_vis is only available for Unix operating systems")



#!/usr/bin/env python3

NAME = "D"
ELEMENT = "D"
NAME_MP = "Mn"
ELEMENT_MP = "Mn"
OCCUPANCY_MP = -1

_pdb_format = "{:6}{:5d}{:>5}{:1}{:>3} {:1}{:4d}{:1}   {:>8.3f}{:>8.3f}{:>8.3f}{:>6.2f}{:>6.2f}          {:>2}{:>2}\n"


def open_dx(grid_3d, xyz_min, d_xyz):
    """Return an OpenDX formatted string

    Parameters
    ----------
    density : ndarray
        3-dimensional array of grid points with a shape given by n_xyz
    xyz_min : list
        origin coordinates of the grid
    d_xyz : list
        grid spacing in x-,y- and z-direction

    Returns
    -------
    str
        OpenDX formatted string

    """
    xmin, ymin, zmin = xyz_min
    nx, ny, nz = grid_3d.shape
    dx, dy, dz = d_xyz

    s = ""
    s += "object 1 class gridpositions counts {:d} {:d} {:d}\n".format(nx, ny, nz)
    s += "origin {:0.3f} {:0.3f} {:0.3f}\n".format(xmin, ymin, zmin)
    s += "delta {:0.2f} 0 0\n".format(dx)
    s += "delta 0 {:0.2f} 0\n".format(dy)
    s += "delta 0 0 {:0.2f}\n".format(dz)
    s += "object 2 class gridconnections counts {:d} {:d} {:d}\n".format(nx, ny, nz)
    s += "object 3 class array type double rank 0 items {:d} data follows\n".format(nx * ny * nz)
    k = 0
    for ix in range(0, nx):
        for iy in range(0, ny):
            for iz in range(0, nz):
                s += "{}".format(grid_3d[ix, iy, iz])
                k += 1
                if k % 3 == 0:
                    s += "\n"
                else:
                    s += " "
    s += 'attribute "dep" string "positions"\n'
    s += 'object "density" class field\n'
    s += 'component "positions" value 1\n'
    s += 'component "connections" value 2\n'
    s += 'component "data" value 3'

    return s


def xyz(cloud_xyzqt, mp, write_weights=True, encode_element=False):
    """Return an XYZ formatted string

    Parameters
    ----------
    cloud_xyzq : ndarray
        array of x-,y-,z-coordinates and corresponding weights with a shape [n_gridpts(+), 4]
    mp : ndarray
         mean position
    write_weights : bool, optional=True
        include weights in XYZ file (5th column)
    encode_element : bool, optional=False

    Returns
    -------
    str
        XYZ formatted string
    """
    n_points = cloud_xyzqt.shape[0]
    s = ""
    s += "{:d}\n".format(n_points)
    s += "# Accessible contact volume\n"
    if write_weights:
        for k in range(n_points):
            if encode_element:
                if cloud_xyzqt[k, 4] == 2:
                    element = "C"  # contact volume
                else:
                    element = "F"  # free volume
            else:
                element = ELEMENT
            s += f"{element:1}\t{cloud_xyzqt[k, 0]:.3f}\t{cloud_xyzqt[k, 1]:.3f}\t{cloud_xyzqt[k, 2]:.3f}\t{cloud_xyzqt[k, 3]:.3f}\n"
        s += f"{ELEMENT_MP:1}\t{mp[0]:.3f}\t{mp[1]:.3f}\t{mp[2]:.3f}\t{-1:.3f}\n"

    else:
        for k in range(n_points):
            if encode_element:
                if cloud_xyzqt[k, 4] == 2:
                    element = "C"
                else:
                    element = "F"
            else:
                element = ELEMENT
            s += f"{element:1}\t{cloud_xyzqt[k, 0]:.3f}\t{cloud_xyzqt[k, 1]:.3f}\t{cloud_xyzqt[k, 2]:.3f}\n"
        s += f"{ELEMENT_MP:1}\t{mp[0]:.3f}\t{mp[1]:.3f}\t{mp[2]:.3f}\n"
    return s


def pdb(cloud_xyzqt, mp):
    """Returns a PDB formatted string

    Parameters
    ----------
    cloud_xyzq : ndarray
        array of x-,y-,z-coordinates and corresponding weights with a shape [n_gridpts(+), 4]
    tag : ndarray
        one-dimensional array of length n_gridpts

    Returns
    -------
    str
        PDB formatted string

    Notes
    -----
    the PDB properties are as follows:

    +--------------------------+------------+-----------+
    | entry                    | PyMOL      | pos.      |
    +==========================+============+===========+
    | ATOM / HETATM            |            | [1-6]     |
    +--------------------------+------------+-----------+
    | atom serial number       | (id)       | [7-11]    |
    +--------------------------+------------+-----------+
    | atom name                | (name)     | [12-16]   |
    +--------------------------+------------+-----------+
    | alternate loc. indicator | (alt)      | [17]      |
    +--------------------------+------------+-----------+
    | residue name             | (resn)     | [18-20]   |
    +--------------------------+------------+-----------+
    |                          |            | [21]      |
    +--------------------------+------------+-----------+
    | chain identifier         | (chain)    | [22]      |
    +--------------------------+------------+-----------+
    | residue seq. number      | (resi)     | [23-26]   |
    +--------------------------+------------+-----------+
    | residue insertion code   |            | [27]      |
    +--------------------------+------------+-----------+
    |                          |            | [28-30]   |
    +--------------------------+------------+-----------+
    | x coordinate (in A)      |            | [31-38]   |
    +--------------------------+------------+-----------+
    | y coordinate (in A)      |            | [39-46]   |
    +--------------------------+------------+-----------+
    | z coordinate (in A)      |            | [47-54]   |
    +--------------------------+------------+-----------+
    | occupancy                | (q)        | [55-60]   |
    +--------------------------+------------+-----------+
    | temperature factor       | (b)        | [61-66]   |
    +--------------------------+------------+-----------+
    |                          |            | [67-66]   |
    +--------------------------+------------+-----------+
    | element                  | (element)  | [77-78]   |
    +--------------------------+------------+-----------+
    | charge                   | (fc)       | [79-80]   |
    +--------------------------+------------+-----------+

    """
    n_points = cloud_xyzqt.shape[0]
    bfactor = 99
    s = ""
    for k in range(n_points):
        if cloud_xyzqt[k, 4] == 2:
            resn = "CV"
        else:
            resn = "FV"
        s += _pdb_format.format(
            "ATOM",
            k + 1,
            NAME,
            " ",
            resn,
            " ",
            int(cloud_xyzqt[k, 4]),
            " ",
            cloud_xyzqt[k, 0],
            cloud_xyzqt[k, 1],
            cloud_xyzqt[k, 2],
            bfactor,
            cloud_xyzqt[k, 3],
            ELEMENT,
            " ",
        )
    s += _pdb_format.format(
        "ATOM", k + 2, NAME_MP, " ", "MP", " ", 0, " ", mp[0], mp[1], mp[2], bfactor, -1, ELEMENT_MP, " "
    )
    return s



#!/usr/bin/env python3

import numpy as np
import os
import argparse
import json
import glob
import multiprocessing
import tqdm
import tqdm.notebook
import itertools
import pickle
import copy
import pandas as pd
import jsonschema
import sys

from . import relaxation
from fretraj import metadata

package_directory = os.path.dirname(os.path.abspath(__file__))


_relax_types = ["D_f", "D_ic", "A_f", "A_ic", "q"]

_schema = {
    "title": "Burst",
    "description": "Parameters to calculate FRET efficiencies and anisotropy decays from MD simulations",
    "type": "object",
    "properties": {
        "dyes": {
            "description": "dye lifetimes and quantum yields",
            "type": "object",
            "properties": {
                "tauD": {"type": "number"},
                "tauA": {"type": "number"},
                "QD": {"type": "number"},
                "QA": {"type": "number"},
                "dipole_angle_abs_em": {"type": "number"},
            },
            "required": ["tauD", "tauA", "QD", "QA"],
        },
        "sampling": {
            "description": "settings for Monte-Carlo sampling",
            "type": "object",
            "properties": {
                "nbursts": {"type": "integer"},
                "skipframesatstart": {"type": "integer"},
                "skipframesatend": {"type": "integer"},
                "multiprocessing": {"type": "boolean"},
            },
            "required": [],
        },
        "fret": {
            "description": "FRET pair parameters",
            "type": "object",
            "properties": {
                "R0": {"type": "number"},
                "kappasquare": {"type": "number"},
                "no_gamma": {"type": "boolean"},
                "quenching_radius": {"type": "number"},
            },
            "required": ["R0"],
        },
        "species": {
            "description": "FRET pair parameters",
            "type": "object",
            "properties": {
                "name": {"type": "array"},
                "unix_pattern_rkappa": {"type": "array"},
                "unix_pattern_don_coords": {"type": "array"},
                "unix_pattern_acc_coords": {"type": "array"},
                "probability": {"type": "array"},
                "n_trajectory_splits": {"type": ["integer", "null"], "minimum": 1},
            },
            "required": ["name", "unix_pattern_rkappa", "probability"],
        },
        "bursts": {
            "description": "FRET pair parameters",
            "type": "object",
            "properties": {
                "lower_limit": {"type": "integer"},
                "upper_limit": {"type": "integer"},
                "lambda": {"type": "number", "maximum": 0},
                "burst_size_file": {"type": ["string", "null"]},
                "QY_correction": {"type": "boolean"},
                "averaging": {"type": "string"},
            },
            "required": ["lower_limit", "upper_limit", "lambda", "averaging"],
        },
    },
}

_defaults = {
    "dyes": {"dipole_angle_abs_em": 0},
    "sampling": {"nbursts": 2000, "skipframesatstart": 0, "skipframesatend": 1000, "multiprocessing": True},
    "fret": {"kappasquare": 0.6666, "no_gamma": False, "quenching_radius": 1},
    "species": {"n_trajectory_splits": None},
    "bursts": {"QY_correction": False, "burst_size_file": None},
}


def parseCmd():
    """Read the command line arguments

    Returns
    -------
    tuple of str
        path to the directory containing the .dat files (rkappa) and .xvg files(dye coordinates)
        as well as the path to a JSON-formatted parameter file for the burst, FRET and anisotropy calculations
    """
    parser = argparse.ArgumentParser(description="Compute FRET histograms from MD simulations with fluorophore labels")
    parser.add_argument("--version", action="version", version="%(prog)s " + str(metadata["__version__"]))
    parser.add_argument("-d", "--directory", help="Directory with gmx dyecouple output files", required=True)
    parser.add_argument("-p", "--parameters", help="Parameter file (.json)", required=True)
    parser.add_argument("-o", "--output", help="Output file o", required=False)
    args = parser.parse_args()
    directory = args.directory
    parameter_file = args.parameters
    return directory, parameter_file


def readParameters(parameter_file):
    """Read parameters from a JSON file

    Parameters
    ----------
    parameter_file : str
        path to a JSON-formatted parameter file (see tutorial for an example)

    Returns
    -------
    dict
        parameters for the burst, FRET and anisotropy calculations
    """
    with open(parameter_file, "r") as f:
        parameters = json.load(f)
        jsonschema.validate(parameters, schema=_schema)
        for section in _defaults.keys():
            for key, val in _defaults[section].items():
                if key not in parameters[section]:
                    parameters[section][key] = val
    return parameters


def in_notebook():
    """Check if code is run in IPython notebook.
    The variable `__IPYTHON__` is defined in Jupyter or IPython but not a normal Python interpreter.

    Returns
    -------
    bool
        Returns `True` if code is executed from Jupyter/IPython notebook
    """
    try:
        __IPYTHON__
        return True
    except NameError:
        return False


class Ensemble:
    """Instantiate a new ensemble of trajectories from different species

    Parameters
    ----------
    directory : str
        path to the directory containing the .dat files (rkappa) and .xvg files (dye coordinates)
    parameters : dict
        parameters for the burst, FRET and anisotropy calculations (see tutorial for an example)
    compute_anisotropy : bool, optional=False
        Calculate the time-resolved ensemble anisotropy. Requires a donor and acceptor .xvg file in the directory.
        Each file should contain xyz-coordinates of two atoms which define the transition dipole of the respective dye.
    verbose : bool, optional=True
        Print status and results to the command line
    """

    def __init__(self, directory, parameters, compute_anisotropy=False, verbose=True, units="A"):
        self.species = []
        ps = parameters["species"]
        for i in range(len(ps["name"])):
            filelist_rkappa = sorted(glob.glob(os.path.join(directory, ps["unix_pattern_rkappa"][i])))
            if not filelist_rkappa:
                raise IndexError(
                    'No rkappa files found for species "{}" with the pattern "{}" in the directory "{}".'.format(
                        ps["name"][i], ps["unix_pattern_rkappa"][i], directory
                    )
                )

            if compute_anisotropy:
                filelist_don_coords = sorted(glob.glob(os.path.join(directory, ps["unix_pattern_don_coords"][i])))
                filelist_acc_coords = sorted(glob.glob(os.path.join(directory, ps["unix_pattern_acc_coords"][i])))
                if (not filelist_don_coords) or (not filelist_acc_coords):
                    raise NameError(
                        'No dye coordinates files found for species "{}" with the pattern "{}" or "{}" in the directory "{}".'.format(
                            ps["name"][i], ps["unix_pattern_don_coords"][i], ps["unix_pattern_acc_coords"][i], directory
                        )
                    )
                elif (len(filelist_rkappa) != len(filelist_don_coords)) or (
                    len(filelist_rkappa) != len(filelist_acc_coords)
                ):
                    raise ValueError(
                        'There are not the same number of rkappa and dye coordinates files in the directory "{}".'.format(
                            directory
                        )
                    )
                else:
                    if verbose:
                        print("Loading files...")
                    self.species.append(
                        Species(
                            ps["name"][i],
                            ps["probability"][i],
                            filelist_rkappa,
                            ps["n_trajectory_splits"],
                            filelist_don_coords,
                            filelist_acc_coords,
                            units=units,
                        )
                    )
            else:
                if verbose:
                    print("Loading files...")
                self.species.append(
                    Species(
                        ps["name"][i], ps["probability"][i], filelist_rkappa, ps["n_trajectory_splits"], units=units
                    )
                )

        self.checkTimeStepIdentity()

    def checkTimeStepIdentity(self):
        """Verify that the trajectories of all species in the ensemble have the same time step"""
        dt = [traj.dt for species in self.species for traj in species.trajectories]
        if len(dt) > 1 and (not dt[1:] == dt[:-1]):
            raise ValueError("Timestep of trajectories is not the same across the ensemble.")
        else:
            self.dt = dt[0]


class Species:
    """Create a new species of trajectories

    Parameters
    ----------
    name : str
        name of the species
    probability : float
        relative weight of the species in the ensemble
    filelist_rkappa : list of str
        list of .dat filenames with inter-dye distance (R) and kappasquare values
    n_trajectory_splits : int, optional=None
        split the trajectory into `n` parts. Together with `averaging="trajectory"` this simulates non-interconverting
        species which leads to broadening; usually this should be left at the default: `None`
    filelist_don_coords, filelist_acc_coords : list of str, optional=None
        list of .xvg filenames with xyz coordinates of two atoms defining
        the transition dipole of the donor or acceptor dye
    """

    def __init__(
        self,
        name,
        probability,
        filelist_rkappa,
        n_trajectory_splits=None,
        filelist_don_coords=None,
        filelist_acc_coords=None,
        units="A",
    ):
        self.name = name
        self.probability = probability
        self.trajectories = []
        total_frames = 0
        for i, rkappa_filename in enumerate(filelist_rkappa):
            if filelist_don_coords and filelist_acc_coords:
                traj = Trajectory.from_file(
                    rkappa_filename, filelist_don_coords[i], filelist_acc_coords[i], units=units
                )
            else:
                traj = Trajectory.from_file(rkappa_filename, units=units)

            if n_trajectory_splits:
                time = np.array_split(traj.time, n_trajectory_splits)
                R = np.array_split(traj.R, n_trajectory_splits)
                kappasquare = np.array_split(traj.kappasquare, n_trajectory_splits)
                if traj.donor_xyz is not None:
                    donor_xyz = np.array_split(traj.donor_xyz, n_trajectory_splits)
                else:
                    donor_xyz = [None] * n_trajectory_splits
                if traj.acceptor_xyz is not None:
                    acceptor_xyz = np.array_split(traj.acceptor_xyz, n_trajectory_splits)
                else:
                    acceptor_xyz = [None] * n_trajectory_splits
                for i in range(n_trajectory_splits):
                    time[i] = np.array([k * (time[i][1] - time[i][0]) for k in range(len(time[i]))])
                    traj = Trajectory(time[i], R[i], kappasquare[i], donor_xyz[i], acceptor_xyz[i], units=units)
                    self.trajectories.append(traj)
                    total_frames += len(traj.time)
            else:
                self.trajectories.append(traj)
                total_frames += len(traj.time)

        for traj in self.trajectories:
            setattr(traj, "weight", len(traj.time) / total_frames)


class Trajectory:
    """Time trajectory (in ps) of inter-dye distances (in nm) and kappa-square values

    Parameters
    ----------
    time : array
        time in picoseconds
    R : array
        inter-dye distance in nanometers
    kappasquare : array
        kappasquare values calculated from the orientation of the donor and acceptor transition dipoles
    donor_xyz, acceptor_xyz: ndarray, optional=None
        xyz coordinates of two atoms defining the transition dipole of the donor or the acceptor dye
    """

    def __init__(self, time, R, kappasquare, donor_xyz=None, acceptor_xyz=None, units="A"):
        self.time = time
        if units == "nm":
            self.R = R * 10
        else:
            self.R = R
        self.kappasquare = kappasquare
        self.length = len(time)
        self.dt = (time[1] - time[0]) / 1000  # in nanoseconds
        self.k_fret = None
        self.p_fret = None
        self.pD_totfret = None
        self.weight = None
        self.donor_xyz = donor_xyz
        self.acceptor_xyz = acceptor_xyz
        if (donor_xyz is not None) and (acceptor_xyz is not None):
            self.checkLengthIdentity(self.length, donor_xyz, acceptor_xyz)
        self.donorTD = self.transitionDipole(donor_xyz)
        self.acceptorTD = self.transitionDipole(acceptor_xyz)

    @classmethod
    def from_file(cls, rkappa_filename, don_coords_filename=None, acc_coords_filename=None, units="A"):
        """Create a burst.Trajectory class from filenames

        Parameters
        ----------
        rkappa_filename : str
            name of a .dat file containing inter-dye distances and kappasquare values
        don_coords_filename, acc_coords_filename : str, optional=None
            name of an .xvg file containing xyz-coordinates of two atoms defining
            the transition dipole of the donor or acceptor dye respectively
        units : {'A', 'nm'}, optional='A'
            distance units ('A': Angstroms, 'nm': nanometers)
        """
        rkappa = pd.read_csv(rkappa_filename, sep=r"\s+", names=["time", "R", "kappa"]).values
        if units == "nm":
            rkappa = rkappa * 10
        if don_coords_filename is not None:
            donor_xyz = pd.read_csv(
                don_coords_filename,
                sep=r"\s+",
                comment="@",
                skiprows=13,
                names=["time", "x1", "y1", "z1", "x2", "y2", "z2"],
            ).values
            if units == "nm":
                donor_xyz[:, 1:] = donor_xyz[:, 1:] * 10
        else:
            donor_xyz = None
        if acc_coords_filename is not None:
            acceptor_xyz = pd.read_csv(
                acc_coords_filename,
                sep=r"\s+",
                comment="@",
                skiprows=13,
                names=["time", "x1", "y1", "z1", "x2", "y2", "z2"],
            ).values
            if units == "nm":
                acceptor_xyz[:, 1:] = acceptor_xyz[:, 1:] * 10
        else:
            acceptor_xyz = None
        return cls(rkappa[:, 0], rkappa[:, 1], rkappa[:, 2], donor_xyz, acceptor_xyz)

    def checkLengthIdentity(self, traj_length, donor_xyz, acceptor_xyz):
        """Check that rkappa and dye coordinate files have the same length

        Parameters
        ----------
        traj_length : int
            number of frames in the trajectory
        donor_xyz, acceptor_xyz: ndarray, optional=None
            xyz coordinates of two atoms defining the transition dipole of the donor or the acceptor dye
        """
        all_lengths = [traj_length, len(donor_xyz[:, 0]), len(acceptor_xyz[:, 0])]
        if all_lengths[1:] != all_lengths[:-1]:
            raise ValueError(
                f"Length of rkappa and dye coordinates is not the same:\nrkappa: {all_lengths[0]}, donor coordinates: {all_lengths[1]}, acceptor coordinates: {all_lengths[2]}".format(
                    all_lengths
                )
            )

    @staticmethod
    def transitionDipole(dye_xyz):
        """Compute a normalized transition dipole vector

        Parameters
        ----------
        dye_xyz : numpy.ndarray
            array of shape [n,7] with columns: `time`, `x1`, `y1`, `z1`, `x2`, `y2`, `z2` where 1 and 2 are two atoms
            defining the transition dipole
        """
        try:
            TD = dye_xyz[:, 4:7] - dye_xyz[:, 1:4]
        except IndexError:
            print(
                'Dye coordinate file has a wrong format. Rerun "gmx traj" with an index file containing two atoms which define the transition dipole.'
            )
            return None
        except TypeError:
            return None
        else:
            return TD / np.linalg.norm(TD, axis=1, keepdims=True)


class Burst:
    """Create a new burst

    Parameters
    ----------
    burstsize : int
        total number of photons (donor + acceptor) to generate in this burst
    QD : float
         donor fluorescence quantum yield
    QA : float
         acceptor fluorescence quantum yield
    """

    def __init__(self, burstsize, QD, QA):
        self.burstsize = burstsize
        self.events_DD_DA = {t: 0 for t in _relax_types}
        self.events_AA = {t: 0 for t in _relax_types}
        self.decaytimes_DD_DA = {t: [] for t in _relax_types}
        self.decaytimes_AA = {t: [] for t in _relax_types}
        self.polarizations = {t: [] for t in ["D_f", "A_f"]}
        self.FRETefficiency = None

    def checkBurstSizeReached(self, QD, QA, QY_correction):
        """Check if the number of photons has reached the specified burstsize

        Parameters
        ----------
        QD : float
            donor fluorescence quantum yield
        QA : float
            acceptor fluorescence quantum yield
        QY_correction : bool
            correct the number of donor and acceptor photons by their respective quantum yield when evaluating whether the burstsize is reached.
            e.g. say we collected 20 donor photons (with QD=0.5) and 25 acceptor photons (with QA=0.75) and the burstsize is set to be 50.
            If QY_correction is `False`, the total photon count is 20+25=45 photons, so the required burstsize is not reached yet.
            If QY_correction is `True`, the total photon count is (20/0.5 + 25/0.75 = 73) and so the burst is complete.
        """
        if QY_correction and (self.events_DD_DA["D_f"] / QD + self.events_DD_DA["A_f"] / QA >= self.burstsize):
            return True
        elif (not QY_correction) and (self.events_DD_DA["D_f"] + self.events_DD_DA["A_f"] == self.burstsize):
            return True

    def addRelaxationEvent(self, event, decaytime, polarization, is_AA):
        """Classify the relaxation event

        Parameters
        ----------
        event : int
            type of relaxation event (1: donor photon, -1: internal conversion from donor,
            2: acceptor photon, -2: internal conversion from acceptor, 0: relaxation due to inter-dye quenching)
        decaytime : float
            time in nanoseconds after the excitation event
        polarization : int
            polarization of the donor or acceptor photon (0: p, parallel, 1: s, orthogonal)
        AA : bool
            If `True`, relaxation event occured after acceptor excitation (as in an ns-ALEX/PIE experiment).
            If `False`, relaxation event occured after donor excitation.
        """
        if is_AA:
            if event == 2:
                self.events_AA["A_f"] += 1
                self.decaytimes_AA["A_f"].append(decaytime)
                self.polarizations["A_f"].append(polarization)
            elif event == -2:
                self.events_AA["A_ic"] += 1
                self.decaytimes_AA["A_ic"].append(decaytime)
        else:
            if event == 1:
                self.events_DD_DA["D_f"] += 1
                self.decaytimes_DD_DA["D_f"].append(decaytime)
                self.polarizations["D_f"].append(polarization)
            elif event == -1:
                self.events_DD_DA["D_ic"] += 1
                self.decaytimes_DD_DA["D_ic"].append(decaytime)
            elif event == 2:
                self.events_DD_DA["A_f"] += 1
                self.decaytimes_DD_DA["A_f"].append(decaytime)
            elif event == -2:
                self.events_DD_DA["A_ic"] += 1
                self.decaytimes_DD_DA["A_ic"].append(decaytime)
            else:
                self.events_DD_DA["q"] += 1
                self.decaytimes_DD_DA["q"].append(decaytime)

    def calcFRET(self, no_gamma, QD, QA):
        """Calculate the transfer efficiency based on the donor and acceptor photon counts upon donor excitation

        Parameters
        ----------
        no_gamma : bool
            mimic an uncorrected FRET experiment (i.e. before gamma-correction) which is affected by the different
            quantum yields of donor and acceptor (note: the detection efficiency ratio is always set to be 1).
            If the simulation should be compared to a gamma-corrected experiment this parameter should be set to `False`.
        QD : float
            donor fluorescence quantum yield
        QA : float
            acceptor fluorescence quantum yield
        """
        if no_gamma:
            self.FRETefficiency = (self.events_DD_DA["A_f"] / QA) / (
                self.events_DD_DA["A_f"] / QA + self.events_DD_DA["D_f"] / QD
            )
        else:
            self.FRETefficiency = self.events_DD_DA["A_f"] / (self.events_DD_DA["A_f"] + self.events_DD_DA["D_f"])


class Relaxation:
    """Create a relaxation event (photon emission or internal conversion)

    Parameters
    ----------
    trajectory : burst.Trajectory
        trajectory object containing inter-dye distances, kappasquare values and donor/acceptor dye transition dipoles
    pD_tot : float
        probability of a donor relaxation event (photon emission + internal conversion)
    pA_tot : float
        probability of a acceptor relaxation event (photon emission + internal conversion)
    quenching_radius : float
        minimum distance beyond which inter-dye quenching will lead to a non-radiative decay (e.g. Dexter energy transfer)
    skipframesatstart : int
        number of frames to skip at the beginning of the trajectory
    skipframesatend : int
        number of frames to skip at the end of the trajectory
    QD : float
        donor fluorescence quantum yield
    QA : float
        acceptor fluorescence quantum yield
    compute_anisotropy : bool
        Calculate the time-resolved ensemble anisotropy.
        Requires a donor and acceptor transition dipole to be present in the trajectory.
    """

    def __init__(
        self,
        trajectory,
        pD_tot,
        pA_tot,
        quenching_radius,
        skipframesatstart,
        skipframesatend,
        QD,
        QA,
        compute_anisotropy,
    ):
        event_DD_DA = 0
        while event_DD_DA == 0:
            self.excitation_ndx_DD_DA = relaxation.findExcitationIndex(
                trajectory.length, skipframesatstart, skipframesatend
            )
            if (self.excitation_ndx_DD_DA < 0) or (self.excitation_ndx_DD_DA > trajectory.length):
                raise ValueError(
                    "The excitation index lies outside of the trajectory. Adjust the skipframesatstart and skipframesatend parameters."
                )
            event_DD_DA, self.relaxation_ndx_DD_DA = relaxation.findRelaxationIndex_DD_DA(
                pD_tot, trajectory.pD_totfret, self.excitation_ndx_DD_DA, trajectory.length
            )  # relaxation upon donor excitation (D emission/IC or FRET)

        event_AA = 0
        while event_AA == 0:
            self.excitation_ndx_AA = relaxation.findExcitationIndex(
                trajectory.length, skipframesatstart, skipframesatend
            )
            if (self.excitation_ndx_AA < 0) or (self.excitation_ndx_AA > trajectory.length):
                raise ValueError(
                    "The excitation index lies outside of the trajectory. Adjust the skipframesatstart and skipframesatend parameters."
                )
            event_AA, self.relaxation_ndx_AA = relaxation.findRelaxationIndex_AA(
                pA_tot, self.excitation_ndx_AA, trajectory.length
            )  # relaxation upon acceptor excitation (A emission/IC)

        self.event_DD_DA = relaxation.checkRelaxationIndex(
            event_DD_DA, self.excitation_ndx_DD_DA, self.relaxation_ndx_DD_DA, trajectory.R, quenching_radius, QD, QA
        )
        self.event_AA = relaxation.checkRelaxationIndex(
            event_AA, self.excitation_ndx_AA, self.relaxation_ndx_AA, trajectory.R, quenching_radius, QD, QA
        )

        if compute_anisotropy and (self.event_DD_DA == 1):
            self.polarization_DD = relaxation.polarization(
                trajectory.donorTD[self.excitation_ndx_DD_DA, :], trajectory.donorTD[self.relaxation_ndx_DD_DA, :]
            )
        else:
            self.polarization_DD = None
        if compute_anisotropy and (self.event_AA == 2):
            self.polarization_AA = relaxation.polarization(
                trajectory.acceptorTD[self.excitation_ndx_AA, :], trajectory.acceptorTD[self.relaxation_ndx_AA, :]
            )
        else:
            self.polarization_AA = None

        self.decaytime_DD_DA = (self.relaxation_ndx_DD_DA - self.excitation_ndx_DD_DA) * trajectory.dt
        self.decaytime_AA = (self.relaxation_ndx_AA - self.excitation_ndx_AA) * trajectory.dt


class Experiment:
    """Setup an in silico fluorescence experiment

    Parameters
    ----------
    directory : str
        path to the directory containing the .dat files (rkappa) and .xvg files(dye coordinates)
    parameters : dict
        parameters for the burst, FRET and anisotropy calculations (see tutorial for an example)
    binwidth : float, optional=0.025
        time between photon bins
    compute_anisotropy : bool, optional=False
        Calculate the time-resolved ensemble anisotropy. Requires a donor and acceptor .xvg file in the directory
        where each contains xyz-coordinates of two atoms which define the transition dipole of the dye.
    verbose : bool, optional=True
        Output status and results to the command line
    show_progress : bool, optional=True
        Display the progress of the burst calculation as a status bar
    """

    def __init__(
        self,
        directory,
        parameters,
        binwidth=0.025,
        compute_anisotropy=False,
        verbose=True,
        show_progress=True,
        units="A",
    ):
        self.parameters = parameters
        self.parameters["fret"]["R0_const"] = parameters["fret"]["R0"] / parameters["fret"]["kappasquare"] ** (1 / 6)
        self.compute_anisotropy = compute_anisotropy
        self.ensemble = Ensemble(directory, parameters, compute_anisotropy, verbose, units=units)
        self.calcTransitionRates()

        if verbose:
            self.print_settings(compute_anisotropy)

        for i, species in enumerate(self.ensemble.species):
            for j, traj in enumerate(species.trajectories):
                self.ensemble.species[i].trajectories[j].k_fret = self.calcTransferRate(traj.R)
                self.ensemble.species[i].trajectories[j].p_fret = (
                    self.ensemble.species[i].trajectories[j].k_fret * self.ensemble.dt
                )
                self.ensemble.species[i].trajectories[j].pD_totfret = (
                    self.ensemble.species[i].trajectories[j].p_fret + self.transProb["pD_tot"]
                )

        if self.parameters["sampling"]["multiprocessing"]:
            burstsizes = self.calcBurstsizes()
            if show_progress:
                with multiprocessing.Pool() as pool:
                    if in_notebook():
                        self.bursts = list(
                            tqdm.notebook.tqdm(
                                pool.imap(
                                    self.calcBurst,
                                    burstsizes,
                                    chunksize=int(self.parameters["sampling"]["nbursts"] / 50),
                                ),
                                total=self.parameters["sampling"]["nbursts"],
                                desc="Calculating bursts",
                                bar_format="{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt} [{remaining} s]",
                            )
                        )
                    else:
                        self.bursts = list(
                            tqdm.tqdm(
                                pool.imap(
                                    self.calcBurst,
                                    burstsizes,
                                    chunksize=int(self.parameters["sampling"]["nbursts"] / 50),
                                ),
                                total=self.parameters["sampling"]["nbursts"],
                                desc="Calculating bursts",
                                bar_format="{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt} [{remaining} s]",
                            )
                        )
            else:
                self.bursts = pool.map(self.calcBurst, burstsizes)
        else:
            n = self.parameters["sampling"]["nbursts"]
            burstsizes = self.calcBurstsizes()
            if show_progress:
                if in_notebook():
                    pbar = tqdm.notebook.tqdm(
                        total=n,
                        desc="Calculating bursts",
                        bar_format="{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt} [{remaining} s]",
                    )
                else:
                    pbar = tqdm.tqdm(
                        total=n,
                        desc="Calculating bursts",
                        bar_format="{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt} [{remaining} s]",
                    )
            self.bursts = []
            for i, bs in enumerate(burstsizes):
                self.bursts.append(self.calcBurst(bs))
                if show_progress:
                    pbar.update()

        if verbose:
            print("Combining burst...")
        self.FRETefficiencies = np.array([burst.FRETefficiency for burst in self.bursts])
        self.burstsizes = np.array([burst.burstsize for burst in self.bursts])
        self.decaytimes_DD_DA = {}
        self.decaytimes_AA = {}
        self.polarizations = {}
        self.polIntensity = {}
        self.anisotropy = {}
        for t in _relax_types:
            self.decaytimes_DD_DA[t] = np.array(
                [
                    decaytime
                    for decaytimes in [burst.decaytimes_DD_DA[t] for burst in self.bursts]
                    for decaytime in decaytimes
                ]
            )
            self.decaytimes_AA[t] = np.array(
                [
                    decaytime
                    for decaytimes in [burst.decaytimes_AA[t] for burst in self.bursts]
                    for decaytime in decaytimes
                ]
            )
        if compute_anisotropy:
            for t in ["D_f", "A_f"]:
                self.polarizations[t] = np.array(
                    [
                        pol
                        for polarizations in [burst.polarizations[t] for burst in self.bursts]
                        for pol in polarizations
                    ]
                )
                if t == "D_f":
                    self.polIntensity[t] = self.polarizationIntensity(
                        binwidth, self.decaytimes_DD_DA[t], self.polarizations[t]
                    )
                else:
                    self.polIntensity[t] = self.polarizationIntensity(
                        binwidth, self.decaytimes_AA[t], self.polarizations[t]
                    )
                self.anisotropy[t] = self.calcAnisotropy(
                    self.polIntensity[t], parameters["dyes"]["dipole_angle_abs_em"]
                )

        if verbose:
            self.print_results()

    @staticmethod
    def polarizationIntensity(binwidth, decaytimes, polarizations):
        """Calculate the polarization-resolved fluorescence intensities

        Parameters
        ----------
        binwidth : float
            time between photon bins
        decaytimes : dict of ndarray
            Decaytimes for the different relaxation types collected from all bursts
        polarizations : dict of ndarray
            Photon polarization for donor photons after donor excitation and acceptor photons after acceptor excitation
            The encoding is 0 = parallel (p) and 1 = orthogonal (s)

        Returns
        -------
        ndarray
            polarization intensity array of shape [n,3] with columns: time bins, p-photons counts (parallel),
            s-photon counts (orthogonal)
        """
        time_pol = sorted(zip(decaytimes, polarizations))
        polIntensity = [[], [], []]
        for key, group in itertools.groupby(time_pol, lambda tp: int(tp[0] // binwidth) + 1):
            lst = list([tp[1] for tp in group])
            s = np.count_nonzero(lst)
            polIntensity[2].append(s)
            polIntensity[1].append(len(lst) - s)
            polIntensity[0].append(key * binwidth)
        return np.array(polIntensity).T

    @staticmethod
    def calcAnisotropy(polIntensity, dipole_angle_abs_em):
        """Calculate the time-resolved anisotropy decay

        Parameters
        ----------
        polIntensity : ndarray
            polarization intensity array of shape [n,3] with columns: time bins, p-photons counts (parallel),
            s-photon counts (orthogonal)
        dipole_angle_abs_em : float
            angle between the absorption and emission dipoles in the absence of depolarization due to rotation

        Returns
        -------
        ndarray
            anisotropy array of shape [n,2] with columns: time bins, anisotropy

        References
        ----------
        .. [1] Lakowicz, Principles of fluorescence spectroscopy, 3rd ed. (2006), pp.355-358.
        """
        loss_photoselection = 2 / 5
        dipole_angle_abs_em_rad = dipole_angle_abs_em / 180 * np.pi
        loss_abs_em_dipoles = (3 * np.cos(dipole_angle_abs_em_rad) ** 2 - 1) / 2
        r = (
            loss_photoselection
            * loss_abs_em_dipoles
            * (polIntensity[:, 1] - polIntensity[:, 2])
            / (polIntensity[:, 1] + 2 * polIntensity[:, 2])
        )
        anisotropy = np.vstack((polIntensity[:, 0], r)).T
        return anisotropy

    def calcTransitionRates(self):
        """Calculate the time-independent transition rates and probabilities"""
        self.rates = {
            "kD_f": self.parameters["dyes"]["QD"] / self.parameters["dyes"]["tauD"],
            "kA_f": self.parameters["dyes"]["QA"] / self.parameters["dyes"]["tauA"],
            "kD_ic": (1 - self.parameters["dyes"]["QD"]) / self.parameters["dyes"]["tauD"],
            "kA_ic": (1 - self.parameters["dyes"]["QA"]) / self.parameters["dyes"]["tauA"],
        }
        self.rates["kD_tot"] = self.rates["kD_f"] + self.rates["kD_ic"]
        self.rates["kA_tot"] = self.rates["kA_f"] + self.rates["kA_ic"]

        self.transProb = {
            "pD_f": self.rates["kD_f"] * self.ensemble.dt,
            "pA_f": self.rates["kA_f"] * self.ensemble.dt,
            "pD_ic": self.rates["kD_ic"] * self.ensemble.dt,
            "pA_ic": self.rates["kA_ic"] * self.ensemble.dt,
            "pD_tot": self.rates["kD_tot"] * self.ensemble.dt,
            "pA_tot": self.rates["kA_tot"] * self.ensemble.dt,
        }

    def calcTransferRate(self, R):
        """Calculated the time-dependent transfer rate

        Parameters
        ----------
        R : ndarray
            inter-dye distances

        Returns
        -------
        ndarray
            Rates of transfer efficiency dependent on the inter-dye distance, the Förster radius and the relative dye orientation (kappasquare)
        """
        k_fret = (
            self.rates["kD_tot"]
            * self.parameters["fret"]["kappasquare"]
            * (self.parameters["fret"]["R0_const"] / R) ** 6
        )
        return k_fret

    def calcBurst(self, burstsize):
        """Compute a photon burst

        Parameters
        ----------
        burstsize : int
            total number of photons (donor + acceptor) to generate in this burst

        Returns
        -------
        burst.Burst
            Burst object containing a series of relaxation events with associated decaytimes, polarizations and FRET efficiencies
        """
        burst = Burst(burstsize, self.parameters["dyes"]["QD"], self.parameters["dyes"]["QA"])
        species_probs = [species.probability for species in self.ensemble.species]
        traj_weights = {
            species.name: [traj.weight for traj in species.trajectories] for species in self.ensemble.species
        }

        if self.parameters["bursts"]["averaging"].lower() != "ensemble":
            species = self.ensemble.species[relaxation.integerChoice(species_probs)]
        if (self.parameters["bursts"]["averaging"].lower() != "ensemble") and (
            self.parameters["bursts"]["averaging"].lower() != "species"
        ):
            trajectory = species.trajectories[relaxation.integerChoice(traj_weights[species.name])]
            self.parameters["bursts"]["averaging"] = "trajectory"

        while True:
            if self.parameters["bursts"]["averaging"].lower() == "ensemble":
                species = self.ensemble.species[relaxation.integerChoice(species_probs)]
            if (self.parameters["bursts"]["averaging"].lower() == "ensemble") or (
                self.parameters["bursts"]["averaging"].lower() == "species"
            ):
                trajectory = species.trajectories[relaxation.integerChoice(traj_weights[species.name])]
            relax = Relaxation(
                trajectory,
                self.transProb["pD_tot"],
                self.transProb["pA_tot"],
                self.parameters["fret"]["quenching_radius"],
                self.parameters["sampling"]["skipframesatstart"],
                self.parameters["sampling"]["skipframesatend"],
                self.parameters["dyes"]["QD"],
                self.parameters["dyes"]["QA"],
                self.compute_anisotropy,
            )
            burst.addRelaxationEvent(relax.event_DD_DA, relax.decaytime_DD_DA, relax.polarization_DD, False)
            burst.addRelaxationEvent(relax.event_AA, relax.decaytime_AA, relax.polarization_AA, True)
            if burst.checkBurstSizeReached(
                self.parameters["dyes"]["QD"], self.parameters["dyes"]["QA"], self.parameters["bursts"]["QY_correction"]
            ):
                break
        burst.calcFRET(
            self.parameters["fret"]["no_gamma"], self.parameters["dyes"]["QD"], self.parameters["dyes"]["QA"]
        )
        return burst

    def calcBurstsizes(self):
        """Compute random burstsizes x_1,x_2,...x_n based on an analytical burstsize distribution P(x) = x^lambda

        Returns
        -------
        ndarray
            array of burstsizes within a range defined by a lower and upper threshold
        """
        if self.parameters["bursts"]["burst_size_file"]:
            exp_burst_sizes = pd.read_csv(
                self.parameters["bursts"]["burst_size_file"], sep=r"\s+", names=["burst_size", "counts"]
            )
            burstsizetable = exp_burst_sizes["burst_size"]
            bcounts = exp_burst_sizes["counts"]
        else:
            burstsizetable = np.arange(
                self.parameters["bursts"]["lower_limit"], self.parameters["bursts"]["upper_limit"] + 1
            )
            bcounts = burstsizetable ** (self.parameters["bursts"]["lambda"])
        rng = np.random.default_rng()
        burstsizes = rng.choice(a=burstsizetable, size=self.parameters["sampling"]["nbursts"], p=bcounts / sum(bcounts))
        return burstsizes

    @staticmethod
    def shot_noise_width(mean_E, N):
        """Compute the width (standard deviation) expected due to shot noise given the total number of photons `N`.
        When `N` is the minimal photon threshold for a burst, i.e. parameters['burst']['lower_limit'],
        then function returns the maximum width expected solely due to shot noise.

        Parameters
        ----------
        mean_E : float
            mean transfer efficiency
        N : int
            total number of photons

        Returns
        -------
        float
            standard deviation of a shot-noise limited Gaussian
        """
        return np.sqrt(mean_E * (1 - mean_E) / N)

    def gaussian(x, mean, sigma):
        """Calculate a Gaussian probability density function

        Parameters
        ----------
        x : array
            independent variable
        mean : float
            expected value
        sigma : float
            standard deviation

        Returns
        -------
        array
            probability density at the given `x`
        """
        return 1 / np.sqrt(2 * np.pi * sigma ** 2) * np.exp(-((x - mean) ** 2) / (2 * sigma ** 2))

    def save(self, filename, remove_bursts=False):
        """Pickle the experiment class to a file

        Parameters
        ----------
        filename : str
            filename for the experiment
        remove_bursts : bool
            remove the bursts from the experiments (strongly reduces the file size)
        """
        with open(filename, "wb") as file:
            if remove_bursts:
                exp_noburst = copy.copy(self)
                exp_noburst.bursts = None
                pickle.dump(exp_noburst, file)
            else:
                pickle.dump(self, file)

    @classmethod
    def load(cls, filename):
        """Load an instance of the experiment class from a pickle file

        Parameters
        ----------
        filename : str
            filename of the experiment
        """
        with open(filename, "rb") as file:
            return pickle.load(file)

    def print_settings(self, compute_anisotropy):
        """Print the rates and settings of the in-silico experiment

        Parameters
        ----------
        compute_anisotropy : bool
            calculate anisotropy decays from the time-dependent orientation of the dye dipole vector
        """
        print("Orientation independent R0_const = {:0.1f} A".format(self.parameters["fret"]["R0_const"]))

        print(
            """
              donor    acceptor
QY            {:0.2f}    {:0.2f}
tau (ns)      {:0.2f}    {:0.2f}
k_f (ns^-1)   {:0.2f}    {:0.2f}
k_ic (ns^-1)  {:0.2f}    {:0.2f}
              """.format(
                self.parameters["dyes"]["QD"],
                self.parameters["dyes"]["QA"],
                self.parameters["dyes"]["tauD"],
                self.parameters["dyes"]["tauA"],
                self.rates["kD_f"],
                self.rates["kA_f"],
                self.rates["kD_ic"],
                self.rates["kA_ic"],
            )
        )

        print("Burst averaging method: {}".format(self.parameters["bursts"]["averaging"]))

        if compute_anisotropy:
            print("Calculate anisotropy: yes\n")
        else:
            print("Calculate anisotropy: no\n")

    def print_results(self):
        print(
            "\n\naverage FRET efficiency: {:0.2f} +- {:0.2f}\n".format(
                np.mean(self.FRETefficiencies), np.std(self.FRETefficiencies)
            )
        )


if __name__ == "__main__":
    directory, parameter_file = parseCmd()
    parameters = readParameters(parameter_file)
    experiment = Experiment(directory, parameters, binwidth=0.025, compute_anisotropy=False, verbose=True, units="A")



import sys

try:
    import importlib.metadata as meta  # Python >=3.8
except ModuleNotFoundError:
    import importlib_metadata as meta  # Python 3.7

PACKAGE = "fretraj"

try:
    metadata = meta.metadata(PACKAGE)
except meta.PackageNotFoundError:
    print(f"{PACKAGE} is not installed yet.")


def _get_urls():
    __urls__ = {}
    try:
        for _url in metadata.get_all("Project-URL"):
            __urls__.update(dict([_url.replace(" ", "").split(",")]))
    except TypeError:
        __urls__["Documentation"] = None
        __urls__["Repository"] = None
    return __urls__


__version__ = metadata["Version"]
__author__ = metadata["Author"]
__keywords__ = metadata["Keywords"]
__license__ = metadata["License"]
__urls__ = _get_urls()


try:
    import LabelLib as ll
except ModuleNotFoundError:
    _LabelLib_found = False
else:
    _LabelLib_found = True

try:
    import nglview
except ModuleNotFoundError:
    _nglview_found = False
    _jupyter_module_available = False
else:
    _nglview_found = True
    _jupyter_module_available = True
    from . import jupyter

try:
    from . import burst
except (ImportError, ModuleNotFoundError):
    _burst_module_available = False
    print("The burst submodule could not be imported\n")
else:
    _burst_module_available = True

from . import cloud
from . import export
from . import fret
from . import grid



#!/usr/bin/env python3

import numpy as np
import pandas as pd
import os
import json
import mdtraj as md
import numba as nb
import copy
import pickle


from fretraj import export
from fretraj import fret
from fretraj import grid
from fretraj import _LabelLib_found

if _LabelLib_found:
    import LabelLib as ll

DISTANCE_SAMPLES = 100000

package_directory = os.path.dirname(os.path.abspath(__file__))

with open(os.path.join(package_directory, "periodic_table.json")) as f:
    _periodic_table = json.load(f)

VDW_RADIUS = dict((key, _periodic_table[key]["vdw_radius"]) for key in _periodic_table.keys())

# set None if the parameter is required
_label_dict = {
    "Position": {
        "pd_key": {
            "attach_id": ((int, float), None),
            "linker_length": ((int, float), None),
            "linker_width": ((int, float), None),
            "dye_radius1": ((int, float), None),
            "dye_radius2": ((int, float), None),
            "dye_radius3": ((int, float), None),
            "cv_thickness": ((int, float), 0),
            "grid_spacing": ((int, float), 1.0),
            "mol_selection": (str, "all"),
            "simulation_type": (str, "AV3"),
            "cv_fraction": ((int, float), 0),
            "state": (int, 1),
            "frame_mdtraj": (int, 0),
            "use_LabelLib": (bool, False),
            "contour_level_AV": ((int, float), 0),
            "contour_level_CV": ((int, float), 0.7),
            "b_factor": (int, 100),
            "gaussian_resolution": (int, 2),
            "grid_buffer": ((int, float), 2.0),
            "transparent_AV": (bool, True),
        }
    },
    "Distance": {"pd_key": {"R0": ((int, float), None), "n_dist": (int, 10 ** 6)}},
}

_label_dict_vals = {
    field: {"pd_key": {key: val[1] for key, val in _label_dict[field]["pd_key"].items()}}
    for field in _label_dict.keys()
}
_default_params = {
    field: {key: val[1] for key, val in _label_dict[field]["pd_key"].items() if val[1] is not None}
    for field in _label_dict.keys()
}


def labeling_params(param_file, verbose=True):
    """Parse the parameter file to get the configuration settings

    Parameters
    ----------
    param_file : str
        json-formatted parameter filename

    Returns
    -------
    dict
        position of labels, dye and grid parameters
    """
    with open(param_file) as f:
        labels_json = json.load(f)

    try:
        check_labels(labels_json, verbose)
    except KeyError as e:
        error_type = e.args[0]
        key = e.args[1]
        pos = e.args[2]
        field = e.args[3]
        print("{}: '{}' in {} {}. Exiting...".format(error_type, key, field, pos))
    except TypeError as e:
        print("Wrong data type: {}".format(e))
    else:
        return labels_json


def check_labels(labels, verbose=True):
    """Check integrity of parameter dictionary

    Parameters
    ----------
    labels : dict
        position of labels, dye and grid parameters
    verbose : bool, optional=True
        be verbose about missing parameter and their fall-back values
    """
    for field in _label_dict.keys():
        if field in labels.keys():
            for pos in labels[field].keys():
                if field == "Position":
                    if "simulation_type" not in labels[field][pos]:
                        labels[field][pos]["simulation_type"] = copy.copy(_default_params[field]["simulation_type"])
                    else:
                        if labels[field][pos]["simulation_type"] == "AV1":
                            labels[field][pos]["dye_radius2"] = 0
                            labels[field][pos]["dye_radius3"] = 0
                    if "state" in labels[field][pos] and "frame_mdtraj" not in labels[field][pos]:
                        labels[field][pos]["frame_mdtraj"] = labels[field][pos]["state"] - 1
                    if "frame_mdtraj" in labels[field][pos] and "state" not in labels[field][pos]:
                        labels[field][pos]["state"] = labels[field][pos]["frame_mdtraj"] + 1
                elif field == "Distance":
                    pass

                # check if all keys that are needed are defined
                for key, (t, _) in _label_dict[field]["pd_key"].items():
                    if key not in labels[field][pos]:
                        if key in _default_params[field].keys():
                            labels[field][pos][key] = copy.copy(_default_params[field][key])
                            if verbose:
                                print(
                                    "Missing Key: '{}' in {} {}. Falling back to \"{}\"".format(
                                        key, field, pos, _default_params[field][key]
                                    )
                                )
                        else:
                            raise KeyError("Missing Key", key, pos, field)
                    else:
                        if not isinstance(labels[field][pos][key], t):
                            raise TypeError(
                                "'{}' in {} {} must be of one of the following types: {}".format(key, field, pos, t)
                            )

                # check if there are any unrecognized keys
                for key in labels[field][pos].keys():
                    if key not in _label_dict_vals[field]["pd_key"].keys():
                        raise KeyError("Unrecognized key", key, pos, field)
        else:
            labels[field] = None
            raise ValueError("Cannot read {} parameters from file: Missing field '{}'.".format(field, field))


def save_obj(filename, obj):
    """
    Save a serialized object to a binary file

    Parameters
    ----------
    filename : str
        filename for pickle object
    obj : serializable object
    """
    with open(filename, "wb") as f:
        try:
            pickle.dump(obj, f)
        except TypeError:
            try:
                obj_tmp = copy.copy(obj)
                obj_tmp.acv = copy.copy(obj.acv)
                del obj_tmp.av
                del obj_tmp.acv.ll_Grid3D
                print("Note: the LabelLib.Grid objects have been removed as they cannot be pickled.")
            except AttributeError:
                print("Error: Cannot pickle the passed object")
            else:
                pickle.dump(obj_tmp, f)


def load_obj(filename):
    """
    Load a serialized object from a binary file

    Parameters
    ----------
    filename : str
        filename of pickle obj
    """
    with open(filename, "rb") as f:
        return pickle.load(f)


def save_labels(filename, labels):
    """Write the ACV parameters to a .json file

    Parameters
    ----------
    filename : str
        filename for label parameters
    labels : dict
        position of labels, dye and grid parameters

    Examples
    --------

    >>> obj.save_labels('parameters.json')

    """
    with open(filename, "w") as f:
        json.dump(labels, f, indent=2)


def printProgressBar(iteration, total, prefix="Progress:", suffix="complete", length=20, fill="█"):
    """Command line progress bar

    Parameters
    ----------
    iteration : int
        current iteration
    total : int
        total number of iterations
    prefix : str, optional='Progress:'
            string before progress bar
    suffix : str, optional='complete'
        string after progress bar
    length : int, optional=20
        length of progress bar
    fill : str, optional='█'
        fill character of progress bar

    Examples
    --------

    >>> printProgressBar(0, n)
    >>> for i in range(n):
            doSomething
            printProgressBar(i + 1, n)

    """
    percent = "{:0.0f}".format(100 * iteration / total)
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + "-" * (length - filledLength)
    print("\r{} |{}| {}% {}".format(prefix, bar, percent, suffix), end="\r")
    if iteration == total:
        print()


def pipeline_frames(structure, donor_site, acceptor_site, labels, frames, fret_pair):
    """Create a pipeline to compute multi-frame donor and acceptor ACVs and calculate a FRET trajectory

    Parameters
    ----------
    structure : mdtraj.Trajectory
                trajectory of atom coordinates loaded from a pdb, xtc or other file
    donor_site : str
                 reference identifier for the donor labeling position
    acceptor_site : str
                    reference identifier for the acceptor labeling position
    labels : dict
             dye, linker and setup parameters for the accessible volume calculation
    frames_mdtraj : list
                    list of frames on the trajectory to be used for the ACV calculation
    fret_pair : str
                Distance key specifying the donor acceptor pair

    Note
    ----
    Running a pipeline more memory-efficient than running Volume.from_frames()
    because the ACVs are stored only until the FRET efficiency is calculated.
    On the downside, no ACV trajectories(.xtc / .xyz) can be saved.
    """
    _labels = copy.copy(labels)
    acv = {}
    fret = []
    n_frames = len(frames)
    printProgressBar(0, n_frames)
    for i, frame in enumerate(frames):
        for dye, site in zip(["D", "A"], [donor_site, acceptor_site]):
            _labels["Position"][site]["frame_mdtraj"] = frame
            _labels["Position"][site]["state"] = frame + 1
            acv[dye] = Volume(structure, site, _labels)
        fret.append(FRET(acv["D"], acv["A"], fret_pair, labels))
        printProgressBar(i + 1, n_frames)
    return fret


def save_mp_traj(filename, volume_list, units="A"):
    """Save a trajectory of dye mean positions as an xyz file

    Parameters
    ----------
    filename : str
        filename for mean position trajectory
    volume_list : array_like
        list of Volume instances
    units : {'A', 'nm'}, optional='A'
        distance units ('A': Angstroms, 'nm': nanometers)
    """
    mps = np.vstack([volume_list[i].acv.mp for i in range(len(volume_list)) if hasattr(volume_list[i].acv, "mp")])
    mps = np.hstack((mps, np.ones((mps.shape[0], 1))))
    mean_mp = np.mean(mps, 0)

    xyz_str = export.xyz(mps, mean_mp, None)
    with open(filename, "w") as fname:
        fname.write(xyz_str)


def acv_subsampling(volume_list, verbose=False):
    """Subsample the ACV to obtain identical number of points over all volumes.
    This allows to create mdtraj.Trajectory objects of the ACV which can be also save as .xtc or .xyz files

    Parameters
    ----------
    volume_list : array_like
        list of Volume instances
    verbose: bool
        be verbose about the number of subsamples
    """
    rng = np.random.default_rng()
    vols = ["AV", "FV", "CV"]
    top = {}
    xyz_sub = {}
    for v in vols:
        if v == "CV":
            xyz_list = list(
                map(lambda volume: volume.acv.cloud_xyzqt[volume.acv.cloud_xyzqt[:, 4] == 2, 0:3], volume_list)
            )
        elif v == "FV":
            xyz_list = list(
                map(lambda volume: volume.acv.cloud_xyzqt[volume.acv.cloud_xyzqt[:, 4] < 2, 0:3], volume_list)
            )
        else:
            xyz_list = list(map(lambda volume: volume.acv.cloud_xyzqt[:, 0:3], volume_list))
        n_pts = min([xyz.shape[0] for xyz in xyz_list])
        top[v] = _create_volume_topology(n_pts)
        xyz_sub[v] = np.stack([rng.choice(xyz, n_pts, replace=False) for xyz in xyz_list])
        if verbose:
            print(f"{n_pts} points sampled in {v}")
    return xyz_sub, top


def create_acv_traj(volume_list, verbose=False):
    """Create a mdtraj.Trajectory object from a volume list

    Parameters
    ----------
    volume_list : array_like
        list of Volume instances
    verbose: bool
        be verbose about the number of subsamples
    """
    acv_traj = {}
    xyz_sub, top = acv_subsampling(volume_list, verbose=False)
    acv_traj["AV"] = md.Trajectory(xyz_sub["AV"] / 10, top["AV"])
    if any(volume_list[0].acv.tag_1d > 1):
        acv_traj["FV"] = md.Trajectory(xyz_sub["FV"] / 10, top["FV"])
        acv_traj["CV"] = md.Trajectory(xyz_sub["CV"] / 10, top["CV"])
    return acv_traj


def save_acv_traj(filename, volume_list, format="pdb", verbose=False):
    """Save a trajectory of ACVs as a full multi model PDB or a subsampled .xyz or .xtc
    (with identical number of points over all volumes)

    Parameters
    ----------
    filename : str
        filename for ACV trajectory
    volume_list : array_like
        list of Volume instances
    format : str
        file format of the ACV (for 'pdb' the full cloud is saved without subsampling;
        for 'xyz' k points in the volumes are randomly picked, where k is the number
        of points in the smallest volume)
    separate_CV : bool
        Save contact and free volume in separate files
        (choose this if you want to retain the CV information)
    """
    if format == "pdb":
        if verbose:
            print('With format "PDB" no subsampling is performed.')
        file_str = ""
        for i, volume in enumerate(volume_list):
            file_str += f"MODEL {i+1}\n"
            file_str += export.pdb(volume.acv.cloud_xyzqt, volume.acv.mp)
            file_str += "ENDMDL\n\n"
        with open(filename, "w") as fname:
            fname.write(file_str)

    elif (format == "xyz") or (format == "xtc"):
        vols = ["AV"]
        if any(volume_list[0].acv.tag_1d > 1):
            vols += ["FV", "CV"]
        base, suffix = os.path.splitext(filename)
        xyz_sub, top = acv_subsampling(volume_list, verbose=False)
        for v in vols:
            filename = f"{base}_{v}{suffix}"
            if format == "xtc":
                with md.formats.XTCTrajectoryFile(filename, "w") as f:
                    f.write(xyz_sub[v] / 10)
            else:
                with md.formats.XYZTrajectoryFile(filename, "w") as f:
                    f.write(xyz_sub[v])

            first_frame = md.Trajectory(xyz_sub[v][0] / 10, top[v])
            first_frame.save_pdb(f"{base}_{v}.pdb")


def load_acv_traj(filename):
    """Load an AV, FV and CV from .xyz or .xtc file

    Parameters
    ----------
    filename : str
        filename of ACV trajectory

    Note
    ----
    Multi model PDB files cannot be loaded as a mdtraj.Trajectory because they are on purpose not subsampled
    and thus contain different number of points per volume
    """
    base, suffix = os.path.splitext(filename)
    acv_traj = {}
    for v in ["FV", "FV", "CV"]:
        acv_traj[v] = md.load(f"{base}_{v}{suffix}", top=f"{base}_{v}.pdb")
    return acv_traj


def _create_volume_topology(n_atoms):
    """Create a topology for an accessible-contact volume

    Parameters
    ----------
    n_atoms : int
        number of points in the volume
    """
    atoms = pd.concat(
        (
            pd.Series(range(n_atoms), name="serial"),
            pd.Series(["D"] * n_atoms, name="name"),
            pd.Series(["D"] * n_atoms, name="element"),
            pd.Series([0] * n_atoms, name="resSeq"),
            pd.Series(["X"] * n_atoms, name="resName"),
            pd.Series([0] * n_atoms, name="chainID"),
        ),
        axis=1,
    )
    top = md.Topology.from_dataframe(atoms)
    return top


def save_structure_traj(filename, structure, frames, format="pdb"):
    """Save selected frames of a trajectory as multi-model PDB or XTC

    Parameters
    ----------
    filename : str
        filename for structure trajectory
    structure : mdtraj.Trajectory
        trajectory of atom coordinates loaded from a pdb, xtc or other file
    format : {'pdb', 'xtc'}
        trajectory file format
    """
    sliced_structure = structure.slice(frames)
    if format == "xtc":
        sliced_structure.save_xtc(filename)
    else:
        sliced_structure.save_pdb(filename)


class ACV:
    """Reweighted accessible contact volume of a covalently linked fluorophore

    Parameters
    ----------
    grid_acv : LabelLib.Grid3D, optional=None
        accessible volume with added weight labels for free and contact volume
        (density label FV: 1.0, density label CV: 2.0); the two volumes are subsequently reweighted
    cv_thickness : float, optional=0
        width of the contact volume in Angstrom (default: 0, i.e. no CV is calculated)
        **Note:** good first approx.: 2*min(dye_radii)
    cv_fraction : float, optional=0
        fraction of dyes that are within the contact volume
        (e.g. as determined by fluorescence anisotropy)
    cloud_xyzq : ndarray, optional=None
        array of x-,y-,z-coordinates and corresponding weights
        with a shape [n_gridpts(+), 4]
    use_LabelLib : bool, optional=True
        make use of LabelLib library to compute FRET values and distances
    """

    def __init__(self, grid_acv=None, cv_thickness=0, cv_fraction=0, cloud_xyzqt=None, use_LabelLib=True):
        if grid_acv is not None:
            self.n_gridpts = np.prod(grid_acv.shape)
            self.grid_1d, self.tag_1d = self._reweight_cv(grid_acv.grid, cv_thickness, cv_fraction)
            self.grid_3d = Volume.reshape_grid(self.grid_1d, grid_acv.shape)
            self.tag_3d = Volume.reshape_grid(self.tag_1d, grid_acv.shape)
            self.cloud_xyzqt = Volume.grid2pts(
                self.grid_3d, grid_acv.originXYZ, np.array([grid_acv.discStep] * 3), self.tag_3d
            )
            if use_LabelLib and _LabelLib_found:
                self.ll_Grid3D = ll.Grid3D(grid_acv.shape, grid_acv.originXYZ, grid_acv.discStep)
                self.ll_Grid3D.grid = self.grid_1d
            else:
                self.ll_Grid3D = None
        else:
            self.cloud_xyzqt = cloud_xyzqt
        self.mp = Volume.mean_pos(self.cloud_xyzqt)

    def _reweight_cv(self, grid, cv_thickness, cv_fraction):
        """Reweight the accessible volume based on contact and free volume

        Parameters
        ----------
        cv_thickness : float
            width of the contact volume in Angstrom (default: 0, i.e. no CV is calculated)
        cv_fraction : float
            relative fraction spent inside the contact volume (i.e. stacked to the surface)

        Returns
        -------
        ndarray
            one-dimensional array of grid points of length n_gridpts

        Notes
        -----
        - A good first approximation of the CV thickness is about two times the smallest dye radius
        - The CV fraction can be calculated from the residual fluorescence anisotropy
        """
        grid_1d = np.array(grid)
        tag_1d = self._tag_volume(grid_1d)
        if cv_thickness > 0:
            weight_cv = self._weight_factor(grid_1d, cv_fraction)
            grid_1d[grid_1d > 1.0] = weight_cv
        grid_1d = np.clip(grid_1d, 0, None)
        return grid_1d, tag_1d

    @staticmethod
    def _weight_factor(grid_1d, cv_fraction):
        """Calculate the weight of contact volume grid points.

        Note
        ----
        The weight factor corresponds to the factor by which the points of the contact volume
        (CV, dye trapped on surface) are favored over those belonging the free volume (FV, free dye diffusion).
        This accounts for the (experimentally) determined fraction of dyes populating the CV.

        Parameters
        ----------
        grid_1d : ndarray
            one-dimensional array of grid points of length n_gridpts

        cv_fraction : float
            fraction of dyes that are within the contact volume

        Returns
        -------
        float
            weight of each grid point that belongs to the contact volume

        """
        n_CV = np.count_nonzero(grid_1d > 1.0)
        if n_CV != 0:
            n_FV = np.count_nonzero([(grid_1d > 0.0) & (grid_1d <= 1.0)])
            if n_FV != 0:
                weight_cv = n_FV / n_CV * cv_fraction / (1.0 - cv_fraction)
            else:
                print("no fraction of free dyes, all stacked")
                weight_cv = 2
        else:
            weight_cv = 1
        return weight_cv

    def _tag_volume(self, grid_1d):
        """Assign a tag to the grid values depending on their location in the cloud (1: free volume, 2: contact volume)

        Parameters
        ----------
        grid_1d : ndarray
            one-dimensional array of grid points length n_gridpts

        Returns
        -------
        ndarray
            one-dimensional array of length n_gridpts
        """
        tag_1d = np.full(self.n_gridpts, 1)
        tag_1d[grid_1d > 1.0] = 2
        return tag_1d


class FRET:
    """FRET class of distance and transfer efficiency for a pair of donor and acceptor ACVs


    Parameters
    ----------
    volume1, volume2 : instances of the Volume class
    fret_pair : str
        Distance key specifying the donor acceptor pair
    labels : dict
        dye, linker and setup parameters for the accessible volume calculation
    R_DA : ndarray
        donor acceptor distance distribution and associate weights (optional)
    verbose : bool

    Examples
    --------

    >>> ft.Molecule(volume1, volume2)

    >>> ft.Molecule(volume1, volume2, use_LabelLib=False)

    """

    def __init__(self, volume1, volume2, fret_pair, labels, R_DA=None, verbose=True):
        try:
            if volume1.acv is None or volume2.acv is None:
                raise TypeError
        except TypeError:
            if verbose:
                print("One accessible volume is empty")
        else:
            self.fret_pair = fret_pair
            self.R0 = labels["Distance"][fret_pair]["R0"]
            self.n_dist = labels["Distance"][fret_pair]["n_dist"]
            self.use_LabelLib = np.all([volume1.use_LabelLib, volume2.use_LabelLib])
            if self.use_LabelLib and _LabelLib_found:
                if R_DA is None:
                    R_DA = fret.dists_DA_ll(volume1.acv, volume2.acv, n_dist=self.n_dist, return_weights=True)
                self.mean_R_DA = fret.mean_dist_DA_ll(volume1.acv, volume2.acv, n_dist=self.n_dist)
                self.sigma_R_DA = fret.std_dist_DA(volume1.acv, volume2.acv, R_DA=R_DA)
                E_DA = fret.FRET_DA(volume1.acv, volume2.acv, R_DA=R_DA, R0=self.R0)
                self.mean_E_DA = fret.mean_FRET_DA_ll(volume1.acv, volume2.acv, R0=self.R0, n_dist=self.n_dist)
                self.sigma_E_DA = fret.std_FRET_DA(volume1.acv, volume2.acv, E_DA=E_DA)
            else:
                if R_DA is None:
                    R_DA = fret.dists_DA(volume1.acv, volume2.acv, n_dist=self.n_dist, return_weights=True)
                self.mean_R_DA = fret.mean_dist_DA(volume1.acv, volume2.acv, R_DA=R_DA)
                self.sigma_R_DA = fret.std_dist_DA(volume1.acv, volume2.acv, R_DA=R_DA)
                E_DA = fret.FRET_DA(volume1.acv, volume2.acv, R_DA=R_DA, R0=self.R0)
                self.mean_E_DA = fret.mean_FRET_DA(volume1.acv, volume2.acv, E_DA=E_DA)
                self.sigma_E_DA = fret.std_FRET_DA(volume1.acv, volume2.acv, E_DA=E_DA)
            self.mean_R_DA_E = fret.mean_dist_DA_fromFRET(
                volume1.acv, volume2.acv, mean_E_DA=self.mean_E_DA, R0=self.R0
            )
            self.sigma_R_DA_E = fret.std_dist_DA_fromFRET(
                volume1.acv, volume2.acv, mean_E_DA=self.mean_E_DA, sigma_E_DA=self.sigma_E_DA, R0=self.R0
            )
            self.R_attach = fret.dist_attach(volume1.attach_xyz, volume2.attach_xyz)
            self.R_mp = fret.dist_mp(volume1.acv, volume2.acv)

    @classmethod
    def from_volumes(cls, volume_list1, volume_list2, fret_pair, labels, R_DA=None):
        """Alternative constructor for the FRET class by reading in a list of donor and acceptor volumes

        Parameters
        ----------
        volume_list1, volume_list2 : array_like
            lists of Volume instances
        fret_pair : str
            identifier for donor acceptor pair
        labels : dict
            dye, linker and setup parameters for the accessible volume calculation
        R_DA : ndarray
            donor acceptor distance distribution and associate weights (optional)
        """
        n_vols1 = len(volume_list1)
        n_vols2 = len(volume_list2)
        try:
            if n_vols1 != n_vols2:
                raise ValueError
        except ValueError:
            print("The length of volume_list1 and volume_list2 is not the same")
        else:
            printProgressBar(0, n_vols1)
            fret = []
            for i in range(n_vols1):
                fret_value = FRET(volume_list1[i], volume_list2[i], fret_pair, labels, R_DA)
                if volume_list1[i].acv is None or volume_list2[i].acv is None:
                    print("Skip list entry {:d}".format(i))
                else:
                    fret.append(fret_value)
                printProgressBar(i + 1, n_vols1)
            return fret

    def save_fret(self, filename):
        """Write the FRET calculation to a json file

        Parameters
        ----------
        filename : str
            filename for FRET prediction

        Examples
        --------

        >>> obj.save_FRET('parameters.json')
        """
        self.values.to_json(filename, indent=2)

    @property
    def values(self):
        """
        Pandas Dataframe of FRET parameters

        Returns
        -------
        pandas.DataFrame
        """
        fret_results = {
            "R0 (A)": (float(f"{self.R0 :0.1f}"), np.nan),
            "<R_DA> (A)": (float(f"{self.mean_R_DA :0.1f}"), float(f"{self.sigma_R_DA :0.1f}")),
            "<E_DA>": (float(f"{self.mean_E_DA :0.2f}"), float(f"{self.sigma_E_DA :0.2f}")),
            "<R_DA_E> (A)": (float(f"{self.mean_R_DA_E :0.1f}"), float(f"{self.sigma_R_DA_E :0.1f}")),
            "R_attach (A)": (float(f"{self.R_attach :0.1f}"), np.nan),
            "R_mp (A)": (float(f"{self.R_mp :0.1f}"), np.nan),
        }
        fret_results = pd.DataFrame(fret_results, index=["value", "std"])
        return fret_results


class Trajectory:
    """Trajectory class of distances and transfer efficiencies from the FRET class

    Parameters
    ----------
    fret : instance of the FRET class
    timestep : int
        time difference between two frames in picoseconds
    """

    def __init__(self, fret, timestep=None, kappasquare=None):
        n = len(fret)
        self.mean_E_DA = np.array([fret[i].mean_E_DA for i in range(n) if hasattr(fret[i], "mean_E_DA")]).round(2)
        self.mean_R_DA = np.array([fret[i].mean_R_DA for i in range(n) if hasattr(fret[i], "mean_R_DA")]).round(1)
        self.mean_R_DA_E = np.array([fret[i].mean_R_DA_E for i in range(n) if hasattr(fret[i], "mean_R_DA_E")]).round(1)
        self.R_attach = np.array([fret[i].R_attach for i in range(n) if hasattr(fret[i], "R_attach")]).round(1)
        self.R_mp = np.array([fret[i].R_mp for i in range(n) if hasattr(fret[i], "R_mp")]).round(1)
        self.timestep = timestep
        self.kappasquare = kappasquare

    @property
    def dataframe(self):
        """Pandas Dataframe view of the Trajectory object

        Returns
        -------
        pandas.DataFrame
            Dataframe with distances and transfer efficiencies of the FRET trajectory
        """
        df = pd.DataFrame(
            (self.mean_R_DA, self.mean_E_DA, self.mean_R_DA_E, self.R_attach, self.R_mp),
            index=["<R_DA> (A)", "<E_DA>", "<R_DA_E> (A)", "R_attach (A)", "R_mp (A)"],
        ).T
        if self.timestep:
            df = pd.concat((df, pd.Series(range(df.shape[0]), name="time (ps)") * self.timestep), axis=1)
        if self.kappasquare:
            df = pd.concat((df, pd.Series(np.ones(df.shape[0]), name="kappasquare") * self.kappasquare), axis=1)
        return df

    def save_traj(self, filename, format="csv", units="A", header=True, R_kappa_only=False):
        """Save the trajectory as a .csv file

        Parameters
        ----------
        filename : str
            filename for .csv trajectory
        format : {'csv', 'txt'}
        units : {'A', 'nm'}, optional='A'
            distance units ('A': Angstroms, 'nm': nanometers)
        header : bool, default=True
            include header in the output file
        R_kappa_only : bool, default=False
            include only time, R_DA and kappasquare columns in the output file (as *gmx dyecoupl*, Gromacs)
        """
        df = self.dataframe
        with open(filename, "w") as f:
            if format == "txt":
                separator = "\t"
            else:
                separator = ","
            if units == "nm":
                df["<R_DA> (A)"] = self.dataframe["<R_DA> (A)"] / 10
                df["<R_DA_E> (A)"] = self.dataframe["<R_DA_E> (A)"] / 10
                df["R_attach (A)"] = self.dataframe["R_attach (A)"] / 10
                df["R_mp (A)"] = self.dataframe["R_mp (A)"] / 10
                df.rename(
                    {header: header.replace("(A)", "(nm)") for header in df.columns.values},
                    axis="columns",
                    inplace=True,
                )
            if R_kappa_only:
                if self.timestep and self.kappasquare:
                    if units == "nm":
                        columns = ["time (ps)", "<R_DA> (nm)", "kappasquare"]
                    else:
                        columns = ["time (ps)", "<R_DA> (A)", "kappasquare"]
                else:
                    raise KeyError("Time or kappasquare is missing in the DataFrame")
            else:
                columns = None
            f.write(df.to_csv(index=False, sep=separator, header=header, columns=columns))


class Volume:
    """Class object holding the accessible contact volume of a specific labeling position

    Parameters
    ----------
    structure : mdtraj.Trajectory
        trajectory of atom coordinates loaded from a pdb, xtc or other file
    site : str
        reference identifier for the labeling position
    labels : dict
        dye, linker and setup parameters for the accessible volume calculation
    cloud_xyzqt : ndarray
        array of x-,y-,z-coordinates with corresponding weights and tags with a shape [n_gridpts(+), 5]
    """

    def __init__(self, structure, site, labels, cloud_xyzqt=None, verbose=True):
        self.structure = structure
        self.attach_id = labels["Position"][site]["attach_id"]

        if self.attach_id is not None:
            self.labeling_site = site
            self.mol_selection = labels["Position"][site]["mol_selection"]
            self.linker_length = labels["Position"][site]["linker_length"]
            self.linker_width = labels["Position"][site]["linker_width"]
            self.simulation_type = labels["Position"][site]["simulation_type"]
            self.dye_radii = np.array(
                [
                    labels["Position"][site]["dye_radius1"],
                    labels["Position"][site]["dye_radius2"],
                    labels["Position"][site]["dye_radius3"],
                ]
            )
            self.grid_spacing = labels["Position"][site]["grid_spacing"]
            self.cv_thickness = labels["Position"][site]["cv_thickness"]
            self.cv_fraction = labels["Position"][site]["cv_fraction"]
            self.state = labels["Position"][site]["state"]
            self.frame_mdtraj = labels["Position"][site]["frame_mdtraj"]
            self.use_LabelLib = labels["Position"][site]["use_LabelLib"]

            try:
                self.n_atoms = self.structure.n_atoms
            except:
                print("{} is no valid mdtraj.Trajectory object".format(self.structure))
            else:
                try:
                    if self.frame_mdtraj != self.state - 1:
                        raise ValueError
                except ValueError:
                    print(
                        "The state {:d} and frame_mdtraj {:d} are not compatible. The frame_mdtraj should be equal to state - 1".format(
                            self.state, self.frame_mdtraj
                        )
                    )
                    self.av = None
                    self.acv = None
                else:
                    try:
                        if self.state > self.structure.n_frames:
                            raise IndexError
                    except IndexError:
                        print(
                            "The state {:d} and mdtraj frame {:d} are out of range, select a state within the range 1 - {:d} or a mdtraj frame within the range 0 - {:d}".format(
                                self.state, self.frame_mdtraj, self.structure.n_frames, self.structure.n_frames - 1
                            )
                        )
                        self.av = None
                        self.acv = None
                    else:
                        try:
                            if self.attach_id < 1 or self.attach_id > self.n_atoms:
                                raise IndexError
                        except IndexError:
                            print(
                                "The attachment position {:d} is out of range, select an index within 1 - {:d}".format(
                                    self.attach_id, self.n_atoms
                                )
                            )
                            self.av = None
                            self.acv = None
                        else:
                            self.attach_id_mdtraj = labels["Position"][site]["attach_id"] - 1
                            self.resi_atom = self.structure.top.atom(self.attach_id_mdtraj)

                            self.av = self.calc_av(self.use_LabelLib)

                            try:
                                if not np.any(np.array(self.av.grid) > 0):
                                    raise ValueError
                            except ValueError:
                                if verbose:
                                    print(
                                        "Empty Accessible volume at position {:d}. Is your attachment point buried?".format(
                                            self.attach_id
                                        )
                                    )
                                self.acv = None
                            else:
                                self.acv = self.calc_acv(self.use_LabelLib)

        elif cloud_xyzqt is not None:
            self.acv = ACV(cloud_xyzqt=cloud_xyzqt)
        else:
            print("Attachment point is unknown")

    @classmethod
    def from_frames(cls, structure, site, labels, frames_mdtraj):
        """Alternative constructor for the Volume class by reading in one or multiple
        frames using the same dye and grid parameters

        Parameters
        ----------
        structure : mdtraj.Trajectory
            trajectory of atom coordinates loaded from a pdb, xtc or other file
        site : str
            reference identifier for the labeling position
        labels : dict
            dye, linker and setup parameters for the accessible volume calculation
        frames_mdtraj : int or list
            list of frames on the trajectory to be used for the ACV calculation

        Returns
        -------
        list
            list of Volume instances
        """
        n_fr = len(frames_mdtraj)
        printProgressBar(0, n_fr)
        multiframe_volumes = []
        _labels = copy.copy(labels)
        for i, frame in enumerate(frames_mdtraj):
            _labels["Position"][site]["frame_mdtraj"] = frame
            _labels["Position"][site]["state"] = frame + 1
            multiframe_volumes.append(cls(structure, site, _labels))
            printProgressBar(i + 1, n_fr)
        return multiframe_volumes

    @classmethod
    def from_attachID(cls, structure, site, labels, attachID):
        """Alternative constructor for the Volume class by reading in one or multiple
        attachment points using the same dye and grid parameters

        Parameters
        ----------
        structure : mdtraj.Trajectory
            trajectory of atom coordinates loaded from a pdb, xtc or other file
        site : str
            reference identifier for the labeling position
        labels : dict
            dye, linker and setup parameters for the accessible volume calculation
        attachID : int or list
            list of attachment ids on the structure to be used for the ACV calculation

        Returns
        -------
        list
            list of Volume instances

        Examples
        --------

        >>> struct = md.load_pdb('data/2m23.pdb')
        >>> labels_global = ft.labeling_params('data/labels_global.json')
        >>> from_attachID(struct, 0, labelsglobals, [2,929])

        """
        n_aID = len(attachID)
        printProgressBar(0, n_aID)
        multisite_volumes = []
        _labels = copy.copy(labels)
        for i, attach_id in enumerate(attachID):
            _labels["Position"][site]["attach_id"] = attach_id
            multisite_volumes.append(cls(structure, site, _labels))
            printProgressBar(i + 1, n_aID)
        return multisite_volumes

    @staticmethod
    def reshape_grid(grid, shape):
        """Convert a 1D-grid to a 3D-Grid

        Parameters
        ----------
        grid : array_like
            one-dimensional grid array/list of length n_gridpts
        shape : list

        Returns
        -------
        ndarray
            3-dimensional array of grid points with a shape given by n_xyz

        Examples
        --------

        >>> grid = numpy.full(27,1)
        >>> ft.cloud.Volume.reshape_grid(grid, (3,3,3))
        array([[[1., 1., 1.],
                [1., 1., 1.],
                [1., 1., 1.]],
                 ...
               [[1., 1., 1.],
                [1., 1., 1.],
                [1., 1., 1.]]])

        """
        grid_3d = np.array(grid, dtype=np.float64).reshape(shape, order="F")
        return grid_3d

    @staticmethod
    @nb.jit(nopython=True)
    def grid2pts(grid_3d, xyz_min, d_xyz, *args):
        """Convert 3D-grid with density values to xyz coordinates with a weight (q)

        Parameters
        ----------
        grid_3d : ndarray([nx,ny,nz])
            3-dimensional array of grid points with a shape given by n_xyz
        xyz_min : ndarray
            origin coordinates of the grid
        d_xyz : ndarray
            grid spacing in x-,y- and z-direction

        Returns
        -------
        ndarray
            array of x-,y-,z-coordinates with corresponding weights and tags with a shape [n_gridpts(+), 5]

        Examples
        --------

        >>> grid_3d = numpy.zeros((3,3,3))
        >>> grid_3d[(1,1,1)] = 1   # FV
        >>> grid_3d[(1,2,1)] = 10  # CV
        >>> ft.cloud.Volume.grid2pts(grid_3d, (0,0,0), (1,1,1))
        array([[ 0.5,  0.5,  0.5,  1.  1],
               [ 0.5,  1. ,  0.5, 10.  1]])

        >>> tag_3d = numpy.zeros((3,3,3))
        >>> grid_3d[(1,1,1)] = 1   # FV
        >>> grid_3d[(1,1,1)] = 2   # CV
        >>> ft.cloud.Volume.grid2pts(grid_3d, (0,0,0), (1,1,1), tag_3d)
        array([[ 0.5,  0.5,  0.5,  1.  1],
               [ 0.5,  1. ,  0.5, 10.  2]])

        """
        xmin, ymin, zmin = xyz_min
        nx, ny, nz = grid_3d.shape
        dx, dy, dz = d_xyz

        if len(args) != 0:
            tag_3d = args[0]
        else:
            tag_3d = np.full(grid_3d.shape, 1)

        cloud_xyzqt = np.empty((nx * ny * nz, 5), dtype=np.float64)

        gdx = np.arange(0, nx, dtype=np.float64) * dx
        gdy = np.arange(0, ny, dtype=np.float64) * dy
        gdz = np.arange(0, nz, dtype=np.float64) * dz

        n = 0
        for ix in range(nx):
            for iy in range(ny):
                for iz in range(nz):
                    d = grid_3d[ix, iy, iz]
                    if d > 0:
                        cloud_xyzqt[n, 0] = gdx[ix] + xmin
                        cloud_xyzqt[n, 1] = gdy[iy] + ymin
                        cloud_xyzqt[n, 2] = gdz[iz] + zmin
                        cloud_xyzqt[n, 3] = d
                        cloud_xyzqt[n, 4] = tag_3d[ix, iy, iz]
                        n += 1

        return cloud_xyzqt[:n]

    @property
    def mol_xyzr(self):
        """Get coordinates and vdW radii of all selected atoms in the structure

        Returns
        -------
        ndarray
            array of x-,y-,z-coordinates and VdW radii with a shape [n_atoms, 4]

        Examples
        --------

        >>> avobj.mol_xyzr

        """
        frame_mdtraj = self.frame_mdtraj
        struct = self.structure
        try:
            radii = np.array([VDW_RADIUS[atom.element.symbol] / 100 for atom in struct.top.atoms], ndmin=2).T
        except KeyError as e:
            print(f"Atom symbol {e} in topology not recognized")
            raise
        else:
            radii[self.attach_id_mdtraj] = 0.0
            try:
                sele = struct.top.select(self.mol_selection)
            except ValueError:
                print(
                    "{} is not a valid expression, please see http://mdtraj.org/latest/atom_selection.html".format(
                        self.mol_selection
                    )
                )
                print('Falling back to "all"')
                sele = struct.top.select("all")
            finally:
                if self.attach_id_mdtraj not in sele:
                    np.append(sele, self.attach_id_mdtraj)
                xyz = struct.xyz[frame_mdtraj] * 10
                xyzr = np.hstack((xyz[sele], radii[sele]))
                return xyzr

    @property
    def attach_xyz(self):
        """Get coordinates of the dye attachment point

        Returns
        -------
        ndarray
            one-dimensional array of x-,y-,z-coordinates of the attachment point

        Examples
        --------

        >>> avobj.attach_xyz

        """
        xyz = self.structure.xyz[self.frame_mdtraj][self.attach_id_mdtraj] * 10
        return xyz.astype(np.float64)

    def save_acv(self, filename, format="pdb", **kwargs):
        """Write accessible contact volume to file

        Parameters
        ----------
        filename : str
            filename for ACV
        format : {'pdb', 'xyz', 'openDX'}
            file format for ACV
        **kwargs : dict
            key value pairs for .xyz files are {write_weights : bool, encode_element : bool}

        Notes
        -----
        - `write_weights` includes the weights of each point as an additional column in the .xyz file
        - `encode_element` uses the element column (first col.) to encode the FV (=F) or CV (=C)

        Examples
        --------

        >>> obj.save_acv('cloud.xyz', format='xyz', write_weights=False)

        """
        if format == "xyz":
            try:
                write_weights = kwargs["write_weights"]
            except KeyError:
                write_weights = True
            try:
                encode_element = kwargs["encode_element"]
            except KeyError:
                encode_element = False
            file_str = export.xyz(self.acv.cloud_xyzqt, self.acv.mp, write_weights, encode_element)
        elif format == "open_dx":
            d_xyz = np.array([self.grid_spacing] * 3)
            xyz_min = self.av.originXYZ
            file_str = export.open_dx(self.acv.grid_3d, xyz_min, d_xyz)
        else:
            file_str = export.pdb(self.acv.cloud_xyzqt, self.acv.mp)
        with open(filename, "w") as fname:
            fname.write(file_str)

    def calc_av(self, use_LabelLib):
        """Calculate the dye accessible volume

        Returns
        -------
        LabelLib.Grid3D
            object containing a 3-dimensional array of grid points and additional attributes (see Notes)

        Notes
        -----
        The attributes of the av LabelLib.Grid3D object are:
            - discStep : float  (the grid spacing)
            - originXYZ : list  (x-/y-/z-coordinate of the grid origin)
            - shape : list      (number of grid points in x-/y-/z-direction)
            - grid : list       (flattened list of grid point values)

        See Also
        --------
        calc_acv : Calculate accessible contact volume

        References
        ----------
        .. [3] Kalinin, S. et al. "A toolkit and benchmark study for FRET-restrained high-precision structural modeling", *Nat. Methods* **9**, 1218–1225 (2012).
        .. [4] Sindbert, S. et al. "Accurate distance determination of nucleic acids via Förster resonance energy transfer: implications of dye linker length and rigidity", *J. Am. Chem. Soc.* **133**, 2463–2480 (2011).

        Examples
        --------

        >>> avobj.calc_av()

        """
        mol_xyzr = self.mol_xyzr
        attach_xyz = self.attach_xyz

        if use_LabelLib and _LabelLib_found:
            if self.simulation_type == "AV1":
                av = ll.dyeDensityAV1(
                    mol_xyzr.T, attach_xyz, self.linker_length, self.linker_width, self.dye_radii[0], self.grid_spacing
                )
            else:
                av = ll.dyeDensityAV3(
                    mol_xyzr.T, attach_xyz, self.linker_length, self.linker_width, self.dye_radii, self.grid_spacing
                )
        else:
            av = grid.Grid3D(
                self.mol_xyzr,
                self.attach_xyz,
                self.linker_length,
                self.linker_width,
                self.dye_radii,
                self.grid_spacing,
                self.simulation_type,
            )
        return av

    def _dye_acc_surf(self):
        """Calculate dye accessible surface by padding the vdW radius with the thickness of the contact volume

        Returns
        -------
        ndarray
            array with dimensions (5,n_atoms) of marked coordinates and padded vdW radii
            (n_atoms = number of atoms in mdtraj.Trajectory)
        """
        cv_label = 2.0
        das_marker = np.full(self.mol_xyzr.shape[0], cv_label)
        das_xyzrm = np.vstack([self.mol_xyzr.T, das_marker])
        das_xyzrm[3] += self.cv_thickness
        return das_xyzrm

    def calc_acv(self, use_LabelLib):
        """Partition the accessible volume into a free and a contact volume

        Returns
        -------
        acv : fretraj.cloud.ACV

        Notes
        -----
        The attributes from the LabelLib.Grid3D object are:
            - discStep : float  (the grid spacing)
            - originXYZ : list  (x-/y-/z-coordinate of the grid origin)
            - shape : list      (number of grid points in x-/y-/z-direction)
            - grid : list       (flattened list of grid point values)

        Additional attributes are:
            - grid_3d : ndarray([shape])
            - n_gridpts : int
            - tag : ndarray([1,n_gridpts])

        See Also
        --------
        calc_av : Calculate accessible volume

        References
        ----------
        .. [1] Steffen, F. D., Sigel, R. K. O. & Börner, R. "An atomistic view on carbocyanine photophysics in the realm of RNA", *Phys. Chem. Chem. Phys.* **18**, 29045–29055 (2016).
        .. [2] Dimura, M. et al. Quantitative FRET studies and integrative modeling unravel the structure and dynamics of biomolecular systems", *Curr. Opin. Struct. Biol.* **40**, 163–185 (2016).

        Examples
        --------

        >>> avobj.calc_acv()

        """
        das_xyzrm = self._dye_acc_surf()
        if use_LabelLib and _LabelLib_found:
            grid_acv = ll.addWeights(self.av, das_xyzrm)
        else:
            self.av.grid_3d = self.av.addWeights(das_xyzrm.T)
            self.av.grid = self.av.grid_3d.flatten(order="F")
            grid_acv = self.av
        acv = ACV(grid_acv, self.cv_thickness, self.cv_fraction, use_LabelLib=use_LabelLib)
        return acv

    @staticmethod
    def mean_pos(cloud_xyzqt):
        """Calculate mean dye position of the accessible contact volume

        Parameters
        ----------
        cloud_xyzq : ndarray
            array of x-,y-,z-coordinates and corresponding weights with a shape [n_gridpts(+), 4]

        Returns
        -------
        ndarray
             mean position
        """
        x = np.dot(cloud_xyzqt[:, 0], cloud_xyzqt[:, 3])
        y = np.dot(cloud_xyzqt[:, 1], cloud_xyzqt[:, 3])
        z = np.dot(cloud_xyzqt[:, 2], cloud_xyzqt[:, 3])
        mp = np.array((x, y, z)) / cloud_xyzqt[:, 3].sum()
        return mp

    @staticmethod
    def _weighted_quantile_1D(arr_1D, weights, quantile):
        """Compute the weighted quantile of a 1D-array

        Parameters
        ----------
        arr_1D : ndarray
            one-dimensional array
        weights : ndarray
            weight of each element of the array
        quantile : float
            quantile to compute (median=0.5)

        Returns
        -------
        quantile : float
        """
        if (quantile > 1) or (quantile < 0):
            raise ValueError("Quantiles must be in the range [0, 1]")

        idx_sorted = np.argsort(arr_1D)
        arr_1D_sorted = arr_1D[idx_sorted]
        weights_sorted = weights[idx_sorted]
        weights_cs = np.cumsum(weights_sorted)
        xp = (weights_cs - (1 - quantile) * weights_sorted) / weights_cs[-1]
        return np.interp(quantile, xp, arr_1D_sorted)

    def save_mp(self, filename, format="plain", units="A"):
        """Write mean dye position to a file

        Parameters
        ----------
        filename : str
            filename for the mean position
        format : {'plain', 'json'}
            file format (plain .txt or .json)
        units : {'A', 'nm'}, optional='A'
            distance units ('A': Angstroms, 'nm': nanometers)

        Examples
        --------
        >>> avobj.save_mp('mp.dat')

        """
        if units == "nm":
            mp = self.acv.mp / 10
        else:
            mp = self.acv.mp
        with open(filename, "w") as f:
            if format == "json":
                mp_dict = {"x": round(mp[0], 3), "y": round(mp[1], 3), "z": round(mp[2], 3)}
                json.dump(mp_dict, f)
            else:
                f.write("{:0.3f}   {:0.3f}   {:0.3f}\n".format(*mp))



#!/usr/bin/env python3
import numpy as np
import numba as nb
import heapq

_dist_list = np.sqrt(np.array([1, 2, 3, 5, 6]))  # reduction to 74 neighbors


class Grid3D:
    """Class object holding a 3D grid

    Parameters
    ----------
    mol_xyzr : ndarray
        array of x-,y-,z-coordinates and VdW radii with a shape [n_atoms, 4]
    attach_xyz : ndarray
        one-dimensional array of x-,y-,z-coordinates of the attachment point
        (corresponds to the center of the grid)
    linker_length : float
        length of the dye linker in Angstrom
    linker_width : float
        diameter of the dye linker in Angstrom
    dye_radii : ndarray([3,1])
        array of dye radii in Angstrom with shape [3,1]
    grid_spacing : float
        spacing between grid points (in A)

    Notes
    -----
    Attributes of the LabelLib class are:
        - discStep : float  (the grid spacing)
        - originXYZ : numpy.array  (x-/y-/z-coordinate of the grid origin)
        - shape : numpy.array      (number of grid points in x-/y-/z-direction)
        - grid : numpy.array       (flattened list of grid point values)
    """

    def __init__(self, mol_xyzr, attach_xyz, linker_length, linker_width, dye_radii, grid_spacing, simulation_type):
        self.discStep = grid_spacing
        self.attach_xyz = attach_xyz
        self.grid_3d, self.originXYZ, self.originAdj = self.make_grid(attach_xyz, linker_length, grid_spacing)
        self.shape = np.array(self.grid_3d.shape)
        self.halfCubeLength = min(self.shape) * self.discStep * 0.5
        self.grid_3d = self.block_molecule(mol_xyzr, 0.5 * linker_width)
        maxR = 3 * grid_spacing
        maxR_source = linker_width + grid_spacing
        self.grid_3d = self.dijkstra_init(maxR, maxR_source)
        self.grid_3d = self.setAboveTreshold(self.grid_3d, linker_length, -4)
        self.grid_3d = self.setAboveTreshold(self.grid_3d, 0, 1)
        if simulation_type == "AV1":
            dye_radii = [dye_radii[0]]
        self.grid_3d = self.excludeConcentricSpheres(mol_xyzr, dye_radii, 1)
        self.grid = self.grid_3d.flatten(order="F")

    @staticmethod
    @nb.jit(nopython=True)
    def make_grid(attach_xyz, linker_length, grid_spacing):
        """Build a 3D grid around the attachment point

        Parameters
        ----------
        attach_xyz : ndarray
            one-dimensional array of x-,y-,z-coordinates of the attachment point
            (corresponds to the center of the grid)
        linker_length : float
            length of the dye linker in Angstrom
        grid_spacing : float
            spacing between grid points (in A)

        Returns
        -------
        tuple of ndarray
            3-dimensional array of grid points with a shape of 2*ll_padRound+1, minimum of the grid, origin of the grid
        """
        ll_pad = linker_length + 3 * grid_spacing  # pad the linker_length
        ll_padRound = np.ceil(ll_pad / grid_spacing) * grid_spacing  # round to next higher grid value
        xyz_min = attach_xyz - ll_padRound
        originAdj = xyz_min - 0.5 * grid_spacing
        gridptsPerEdge = 2 * int(ll_padRound / grid_spacing + 0.5)  # + 1
        grid_3d = np.full((gridptsPerEdge, gridptsPerEdge, gridptsPerEdge), np.inf)
        for i in range(gridptsPerEdge):
            for j in range(gridptsPerEdge):
                for k in range(gridptsPerEdge):
                    ijk = np.array([i, j, k])
                    maxRSq = int(linker_length ** 2 / grid_spacing ** 2 + 0.5)
                    ijk0 = (attach_xyz - xyz_min) / grid_spacing
                    dSq = np.sum((ijk - ijk0) ** 2)
                    if dSq > maxRSq:
                        grid_3d[(i, j, k)] = -1
        return grid_3d, xyz_min, originAdj

    @staticmethod
    def _xyz2idx(xyz, originAdj, grid_spacing, decimals=6):
        """Get the ijk grid indices for a set of xyz values

        Parameters
        ----------
        xyz : ndarray
            xyz coordinates
        originAdj : ndarray
            origin of the grid
        grid_spacing : float
            spacing between grid points (in A)
        decimals : int
            decimal to round to

        Note
        ----
        Round to n-decimal places (here: 4) before casting to integer
        """
        return np.round(((xyz - originAdj) / grid_spacing), decimals).astype(np.int64)

    def block_molecule(self, mol_xyzr, extraClash):
        """Block the grid points which are within the VdW radius of any atom of the biomolecule plus an extra clash radius

        Parameters
        ----------
        mol_xyzr : ndarray
            array of x-,y-,z-coordinates and VdW radii with a shape [n_atoms, 4]
        extraClash : float
            clash radius added to the VdW radius

        Returns
        -------
        ndarray
            3-dimensional array of grid points with a shape of 2*adjL+1
        """
        maxVdW_extraClash = np.max(mol_xyzr[:, 3]) + extraClash
        neighbor_list = nb.typed.List()
        [neighbor_list.append(n) for n in self.sortedNeighborIdx(maxVdW_extraClash)]
        if not neighbor_list:
            raise ValueError("Neighbor list is empty")
        ijk_atom = self._xyz2idx(mol_xyzr[:, 0:3], self.originAdj, self.discStep)
        outDistSq = (self.halfCubeLength + maxVdW_extraClash) ** 2
        distSq = np.sum((mol_xyzr[:, 0:3] - self.attach_xyz) ** 2, 1)
        grid_3d = self._carve_VdWextraClash(
            self.grid_3d, mol_xyzr, neighbor_list, ijk_atom, extraClash, distSq, outDistSq, self.shape
        )
        return grid_3d

    @staticmethod
    @nb.jit(nopython=True)
    def _carve_VdWextraClash(grid_3d, mol_xyzr, neighbor_list, ijk_atom, extraClash, distSq, outDistSq, grid_shape):
        """Loop through the atoms and assign -1 to all grid values that are within the atoms VdW radius
        plus an extraClash value

        Parameters
        ----------
        grid_3d : ndarray
            3-dimensional array of grid points with a shape of 2*adjL+1
        mol_xyzr : ndarray
            array of x-,y-,z-coordinates and VdW radii with a shape [n_atoms, 4]
        neighbor_list : list of 2-tuples of ndarray and float
            the tuples contain the ijk indices and the distance from the origin (0,0,0)
        ijk_atom : array
            ijk grid indices of the atoms of the biomolecule
        extraClash : float
            clash radius added to the VdW radius
        distSq : float
            square of the distance between the grid points to the attachment point
        outDistSq : float
            square of (halfcubelength + maxVdW_extraClash)
        grid_shape : array
            number of grid points in x,y and z coordinates
        """
        for m in range(mol_xyzr.shape[0]):
            if distSq[m] > outDistSq:
                continue
            for n in neighbor_list:
                if n[0] > mol_xyzr[m, 3] + extraClash:
                    break
                ijk_n = ijk_atom[m] + n[1]
                i, j, k = ijk_n
                if not np.any(ijk_n < 0) and not np.any(grid_shape - ijk_n <= 0):
                    grid_3d[i, j, k] = -1
        return grid_3d

    def dijkstra_init(self, maxR, maxR_source):
        """Initialize the Dijkstra search algorithm

        Parameters
        ----------
        maxR : float
            radius within to search for neighbors
        maxRsource : float
            radius within to search for neighbors in the first round of the Dijkstra algorithm

        Returns
        -------
        ndarray
            3-dimensional array of grid points with a shape of 2*adjL+1
        """
        ai, aj, ak = self._xyz2idx(self.attach_xyz, self.originAdj, self.discStep)
        self.grid_3d[ai, aj, ak] = 0
        priority_queue = [(0.0, (ai, aj, ak))]
        priority_queue = nb.typed.List()
        [priority_queue.append(x) for x in [(0.0, (ai, aj, ak))]]
        edges_ess = nb.typed.List()
        [edges_ess.append(x) for x in self.essential_neighbors(self.sortedNeighborIdx(maxR), _dist_list)]
        edges_src = nb.typed.List()
        [edges_src.append(x) for x in self.sortedNeighborIdx(maxR_source)]
        grid_3d = self.dijkstra(self.grid_3d, edges_ess, edges_src, priority_queue, (ai, aj, ak))
        return grid_3d

    def sortedNeighborIdx(self, maxR):
        """Build a neighbor list with a maximal extent given by maxR and sorted by increasing distance
        from the origin

        Parameters
        ----------
        maxR : float
            radius within to search for neighbors

        Returns
        -------
        list of 2-tuples of ndarray and float
            the tuples contain the ijk indices and the distance from the origin (0,0,0)
        """
        idxs = self._neighborIdx(maxR, self.discStep)
        idxs.sort(key=lambda x: x[0])
        return idxs

    @staticmethod
    def _neighborIdx(maxR, grid_spacing):
        """Build a list of neighboring indices to the origin (0,0,0)

        Parameters
        ----------
        maxR : float
            radius within which to search for neighbors
        grid_spacing : float
            spacing between grid points (in A)

        Returns
        -------
        list of 2-tuples of ndarray and float
            the tuples contain the ijk indices and the distance from the origin (0,0,0), e.g. [(1.0,array([-1,0,0]))]

        Notes
        -----
        The neighbor list is built from an origin set at (0,0,0). imaxR defines the number of indices
        along one axis (i,j or k) to build the index cube.
        For imaxR=3, first a cube of 7*7*7=343 indices is built (1).
        This is reduced in a second step to 123 indices that have a distance < imaxR from the origin (2).
        Finally, they are reduced 74 essential neighbors by the distances in _dist_list (3).
        (this last reduction speeds up the Dikstra algorithm due to the shorter neighbor list without loosing much
        accuracy)
        """
        idxs = []
        maxRSq = maxR ** 2
        imaxR = int(maxR / grid_spacing + 0.5)  # rounds to next integer
        # (1) build a cube with (2*imaxR+1)**3 indices
        for k in range(-imaxR, imaxR + 1, 1):
            for j in range(-imaxR, imaxR + 1, 1):
                for i in range(-imaxR, imaxR + 1, 1):
                    ijk = np.array([i, j, k], dtype=int)
                    # (2) only accept indices with 0 < dS < imaxRSq
                    dSq = np.sum(ijk ** 2) * grid_spacing ** 2
                    if dSq <= maxRSq:  # and dSq > 0:
                        d = np.sqrt(dSq)
                        idxs.append((d, ijk))
        return idxs

    def essential_neighbors(self, idxs, dist_list):
        """Reduce the neighbor list to those indices with a distance from the origin featured in dist_list

        Parameters
        ----------
        idxs : list of 2-tuples of ndarray and float
            the tuples contain the ijk indices and the distance from the origin (0,0,0)
        dist_list : list
            list of distances for neighbor search

        Notes
        -----
        If dist_list=sqrt([1, 2, 3, 5, 6]) then the the neighbor list is reduced to 74 neighbors
        which make up a spherical shape and are a good compromise between neighbor space and time efficiency
        in the Dijkstra algorithm. The smaller this list the faster the Dijkstra algorithm will run
        at the expense of the accuracy of the resulting accessible volume (i.e. its "sphericalness")
        """
        idxs_ess = []
        for e in idxs:
            minDiff = min(abs(e[0] - dist_list * self.discStep))
            if minDiff < (0.01 * self.discStep):
                idxs_ess.append(e)
        return idxs_ess

    @staticmethod
    @nb.jit(nopython=True)
    def dijkstra(grid_3d, edges_ess, edges_src, priority_queue, start_ijk):
        """Djikstra algorithm with a priority queue

        Parameters
        ----------
        grid_3d : ndarray
            3-dimensional array of grid points with a shape given by n_xyz
        edges_ess : nb.typed.List of 2-tuples of ndarray and float
            list n essential neighbors from the origin (0,0,0)
        edges_src : list of 2-tuples of ndarray and float
            nb.typed.List of neighbors in the initialization round of the Dijkstra algorithm
        priority_queue : list of 2-tuples of ndarray and float
            list with distance and index of origin
        start_ijk : array
            ijk grid indices of the attachment site

        Returns
        -------
        ndarray
            3-dimensional array of grid points with a shape given by n_xyz
        """
        priority_queue = list(priority_queue)
        while priority_queue:
            r, idx = heapq.heappop(priority_queue)
            if r > grid_3d[idx]:
                continue
            if idx == start_ijk:
                edges = edges_src
            else:
                edges = edges_ess
            for e in edges:
                i = idx[0] + e[1][0]
                j = idx[1] + e[1][1]
                k = idx[2] + e[1][2]
                val = grid_3d[i, j, k]
                if val < 0:
                    continue
                t_r = r + e[0]
                if t_r < val:
                    grid_3d[i, j, k] = t_r
                    heapq.heappush(priority_queue, (t_r, (i, j, k)))
        return grid_3d

    @staticmethod
    @nb.jit(nopython=True)
    def setAboveTreshold(grid_3d, treshold, new_value):
        """Reassign grid values which are above a specified treshold

        Parameters
        ----------
        grid_3d : ndarray
            3-dimensional array of grid points with a shape given by n_xyz
        treshold : float
            grid values above this treshold are reassigned with new_value
        new_value : float
            new grid value to be assigned

        Returns
        -------
        ndarray
            3-dimensional array of grid points with a shape given by n_xyz
        """
        for i in range(grid_3d.shape[0]):
            for j in range(grid_3d.shape[1]):
                for k in range(grid_3d.shape[2]):
                    if grid_3d[(i, j, k)] > treshold:
                        grid_3d[(i, j, k)] = new_value
        return grid_3d

    def excludeConcentricSpheres(self, mol_xyzr, dye_radii, maxRho):
        """Exclude all grid points which are not compatible with any of the dye radii because of clashing with the VdW surface

        Parameters
        ----------
        mol_xyzr : ndarray
            array of x-,y-,z-coordinates and VdW radii with a shape [n_atoms, 4]
        dye_radii : ndarray([3,1])
            array of dye radii in Angstrom with shape [3,1]
        maxRho : float
            maximum grid value in the free volume

        Returns
        -------
        ndarray
            3-dimensional array of grid points with a shape given by n_xyz
        """
        maxVdW_extraClash = np.max(mol_xyzr[:, 3]) + np.max(dye_radii)
        neighbor_list = nb.typed.List()
        [neighbor_list.append(n) for n in self.sortedNeighborIdx(maxVdW_extraClash)]
        dye_radii_sorted = nb.typed.List()
        [dye_radii_sorted.append(x) for x in sorted(dye_radii)]
        rhos = np.linspace(0, maxRho, len(dye_radii) + 1)
        ijk_atom = self._xyz2idx(mol_xyzr[:, 0:3], self.originAdj, self.discStep)
        outdistSq = (self.halfCubeLength + maxVdW_extraClash) ** 2
        distSq = np.sum((mol_xyzr[:, 0:3] - self.attach_xyz) ** 2, 1)
        grid_3d = self._assignRho(
            self.grid_3d, mol_xyzr, neighbor_list, ijk_atom, dye_radii_sorted, rhos, distSq, outdistSq, self.shape
        )
        return grid_3d

    @staticmethod
    @nb.jit(nopython=True)
    def _assignRho(grid_3d, mol_xyzr, neighbor_list, ijk_atom, dye_radii_sorted, rhos, distSq, outdistSq, grid_shape):
        """Loop through the atoms and radii and reassign the grid values with a number 0 < rho < 1.
        0.00: grid point is not compatible with any of the three dye radii (clashes with VdW surface)
        0.33: grid point is compatible with the smallest dye radius
        0.66: grid point is compatible with the smallest two radii
        1.00: grid point is compatible with all three dye radii

        Parameters
        ----------
        grid_3d : ndarray
            3-dimensional array of grid points with a shape given by n_xyz
        mol_xyzr : ndarray
            array of x-,y-,z-coordinates and VdW radii with a shape [n_atoms, 4]
        neighbor_list : list of 2-tuples of ndarray and float
            the tuples contain the ijk indices and the distance from the origin (0,0,0)
        ijk_atom : array
            ijk grid indices of the atoms of the biomolecule
        dye_radii_sorted : array
            dye radii sorted in ascending order
        rhos : array
            array of grid values for assignment, for AV3 [0, 0.33, 0.66, 1] and for AV1 [0, 1]
        distSq : float
            square of the distance between the grid points to the attachment point
        outdistSq : float
            square of (halfcubelength + maxVdW_extraClash)
        grid_shape : array
            number of grid points in x,y and z coordinates
        """
        n = len(neighbor_list)
        n_radii = len(dye_radii_sorted)
        for m in range(mol_xyzr.shape[0]):
            if distSq[m] > outdistSq:
                continue
            else:
                s = 0
                for r in range(n_radii):
                    effR = dye_radii_sorted[r] + mol_xyzr[m, 3]
                    while neighbor_list[s][0] <= effR:
                        ijk_n = ijk_atom[m] + neighbor_list[s][1]
                        i, j, k = ijk_n
                        if not np.any(ijk_n < 0) and not np.any(grid_shape - ijk_n <= 0):
                            gridval = grid_3d[i, j, k]
                            grid_3d[i, j, k] = min(rhos[r], gridval)
                        s += 1
                        if s >= n:
                            break
        return grid_3d

    def addWeights(self, das_xyzrm):
        """Reassign those grid values which are part of the contact volume.

        Parameters
        ----------
        das_xyzrm : ndarray
            array of xyz coordinates, padded vdW radii and a marking (2.0)

        Returns
        -------
        ndarray
            3-dimensional array of grid points with a shape given by n_xyz
        """
        maxClash = np.max(das_xyzrm[:, 3]) + self.discStep
        neighbor_list = nb.typed.List()
        [neighbor_list.append(n) for n in self.sortedNeighborIdx(maxClash)]
        ijk_atom = self._xyz2idx(das_xyzrm[:, 0:3], self.originAdj, self.discStep)
        outDistSq = (self.halfCubeLength + maxClash) ** 2
        distSq = np.sum((das_xyzrm[:, 0:3] - self.attach_xyz) ** 2, 1)
        grid_3d = self._assignDensity(self.grid_3d, das_xyzrm, neighbor_list, ijk_atom, distSq, outDistSq, self.shape)
        return grid_3d

    @staticmethod
    @nb.jit(nopython=True)
    def _assignDensity(grid_3d, das_xyzrm, neighbor_list, ijk_atom, distSq, outDistSq, grid_shape):
        """Loop through the atom of the biomolecule and reassign those grid value which are within das_xyzrm[m, 3]
        as they belong to the CV.

        Parameters
        ----------
        grid_3d : ndarray
            3-dimensional array of grid points with a shape given by n_xyz
        das_xyzrm : ndarray
            array of marked coordinates and padded vdW radii (n_atoms = number of atoms in mdtraj.Trajectory)
        neighbor_list : list of 2-tuples of ndarray and float
            the tuples contain the ijk indices and the distance from the origin (0,0,0)
        ijk_atom : array
            ijk grid indices of the atoms of the biomolecule
        distSq : float
            square of the distance between the grid points to the attachment point
        outDistSq : float
            square of (halfcubelength + maxVdW_extraClash)
        grid_shape : array
            number of grid points in x,y and z coordinates

        Returns
        -------
        ndarray
            3-dimensional array of grid points with a shape given by n_xyz
        """
        n = len(neighbor_list)
        for m in range(das_xyzrm.shape[0]):
            if distSq[m] > outDistSq:
                continue
            else:
                s = 0
                while neighbor_list[s][0] <= das_xyzrm[m, 3]:
                    ijk_n = ijk_atom[m] + neighbor_list[s][1]
                    i, j, k = ijk_n
                    if not np.any(ijk_n < 0) and not np.any(grid_shape - ijk_n <= 0):
                        gridval = grid_3d[i, j, k]
                        if gridval > 0:
                            grid_3d[i, j, k] += das_xyzrm[m, 4]
                    s += 1
                    if s >= n:
                        break
        return grid_3d
